import dataclasses
from datetime import date
from datetime import timedelta

from django_partitioned_audit.partitions.partition_info import PartitionInfo
from django_partitioned_audit.partitions.partition_info import TimeRange


class TimeRangeGenerator:
    def generate_partition_info_for_current_period(self, today: date) -> PartitionInfo:
        raise NotImplementedError()

    def new_partition(self, previous: TimeRange) -> TimeRange:
        raise NotImplementedError()


#
# Rolling month partitioning
#


@dataclasses.dataclass
class YearMonth:
    year: int
    month: int

    def incr_month(self) -> "YearMonth":
        assert 1 <= self.month <= 12
        if self.month == 12:
            self.year += 1
            self.month = 1
        else:
            self.month += 1
        return self

    def as_date(self, day_of_month=1) -> date:
        assert self.year > 0
        assert 1 <= self.month <= 12
        assert 1 <= day_of_month <= 31
        return date(self.year, self.month, day_of_month)

    @classmethod
    def new_from_date(cls, reference: date) -> "YearMonth":
        return YearMonth(year=reference.year, month=reference.month)


class MonthlyTimeRangeGenerator(TimeRangeGenerator):
    def generate_partition_info_for_current_period(self, today: date) -> TimeRange:
        range_start = YearMonth(today.year, today.month).as_date()
        range_end = YearMonth.new_from_date(today).incr_month().as_date()
        return TimeRange(range_start, range_end)

    def new_partition(self, previous: TimeRange) -> TimeRange:
        new_range_start = previous.to_date
        new_range_end = YearMonth.new_from_date(previous.to_date).incr_month().as_date()
        return TimeRange(new_range_start, new_range_end)


class WeeklyTimeRangeGenerator(TimeRangeGenerator):
    def generate_partition_info_for_current_period(self, today: date) -> TimeRange:
        range_start = today
        range_end = range_start + timedelta(days=7)
        return TimeRange(range_start, range_end)

    def new_partition(self, previous: TimeRange) -> TimeRange:
        new_range_start = previous.to_date
        new_range_end = new_range_start + timedelta(days=7)
        return TimeRange(new_range_start, new_range_end)


class DailyTimeRangeGenerator(TimeRangeGenerator):
    def generate_partition_info_for_current_period(self, today: date) -> TimeRange:
        range_start = today
        range_end = range_start + timedelta(days=1)
        return TimeRange(range_start, range_end)

    def new_partition(self, previous: TimeRange) -> TimeRange:
        new_range_start = previous.to_date
        new_range_end = new_range_start + timedelta(days=1)
        return TimeRange(new_range_start, new_range_end)


class NotImplementedTimeRangeGenerator(TimeRangeGenerator):
    def generate_partition_info_for_current_period(self, today: date) -> TimeRange:
        raise NotImplementedError()

    def new_partition(self, previous: TimeRange) -> TimeRange:
        raise NotImplementedError()



import dataclasses
import re
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch

import pytest

from django_partitioned_audit import signals
from tests.conftest import skipifadvanced


@dataclasses.dataclass
class ModelClass:
    value: str

    @property
    def _meta(self):
        meta = Mock()
        meta.db_table = self.value
        return meta


class SampleAppConfig:
    name = "app_name"
    trigger_audit_models = ["Model1", "Model2"]
    audit_disabled = ["Model3", "Model4"]


@skipifadvanced
@pytest.mark.django_db(transaction=True)
def test_signals():
    with patch("django_partitioned_audit.signals.connections") as mock_connections:
        mock_execute = MagicMock()
        mock_connections.__getitem__.return_value = MagicMock()
        mock_connections.__getitem__.return_value.cursor.return_value = MagicMock()
        mock_connections.__getitem__.return_value.cursor.return_value.__enter__.return_value = MagicMock()
        mock_connections.__getitem__.return_value.cursor.return_value.__enter__.return_value.execute = mock_execute

        app_config = SampleAppConfig()
        app_config.get_model = Mock(side_effect=lambda arg: ModelClass(value=arg))
        signals.audit_run_post_migrate(app_config, verbosity=0, using="default")

        executed_sql_list = [a_call.args[0].lower() for a_call in mock_execute.call_args_list]

        enabled = [
            re.search(
                r"create trigger trigger_audit_entry_creator_trigger after insert or update or delete on (\S+)", sql
            )
            for sql in executed_sql_list
        ]
        enabled = [_.group(1) for _ in enabled if _]

        disabled = [
            re.search(r"drop trigger if exists trigger_audit_entry_creator_trigger on (\S+);", sql)
            for sql in executed_sql_list
        ]
        disabled = [_.group(1) for _ in disabled if _]

        assert set(enabled) == {"model1", "model2"}
        assert set(disabled) == {"model3", "model4"}



import random
from datetime import date

import pytest

from django_partitioned_audit.partitions.partition_info import PartitionInfo
from django_partitioned_audit.partitions.partition_manager_plan import Plan


def test_valid_empty_plan():
    plan = Plan()
    assert plan.sorted_all == []
    assert plan.sorted_existing == []
    assert plan.sorted_to_create == []
    assert plan.last_partition is None
    assert plan.covers_date(date(1000, 1, 1)) is False
    assert plan.covers_date(date(2999, 1, 1)) is False
    plan.validate_existing_partitions()
    plan.validate_to_create()
    plan.validate_all()


def test_valid_plan_with_existing():
    plan = Plan()
    partition = PartitionInfo.create("part", date(1991, 1, 1), date(1991, 2, 1))
    plan.existing_partitions.append(partition)
    assert plan.sorted_all == [partition]
    assert plan.sorted_existing == [partition]
    assert plan.sorted_to_create == []
    assert plan.last_partition == partition
    assert plan.covers_date(date(1990, 12, 31)) is False
    assert plan.covers_date(date(1991, 1, 1))
    assert plan.covers_date(date(1991, 1, 31))
    assert plan.covers_date(date(1991, 2, 1)) is False
    assert plan.covers_date(date(1000, 1, 1)) is False
    assert plan.covers_date(date(2999, 1, 1)) is False
    plan.validate_existing_partitions()
    plan.validate_to_create()
    plan.validate_all()


def test_valid_plan_with_partitions_to_create():
    plan = Plan()
    partition = PartitionInfo.create("part", date(1991, 1, 1), date(1991, 2, 1))
    plan.partitions_to_create.append(partition)
    assert plan.sorted_all == [partition]
    assert plan.sorted_existing == []
    assert plan.sorted_to_create == [partition]
    assert plan.last_partition == partition
    assert plan.covers_date(date(1990, 12, 31)) is False
    assert plan.covers_date(date(1991, 1, 1))
    assert plan.covers_date(date(1991, 1, 31))
    assert plan.covers_date(date(1991, 2, 1)) is False
    assert plan.covers_date(date(1000, 1, 1)) is False
    assert plan.covers_date(date(2999, 1, 1)) is False
    plan.validate_existing_partitions()
    plan.validate_to_create()
    plan.validate_all()


def test_valid_plan_with_both():
    plan = Plan()
    partition_existing = PartitionInfo.create("part", date(1991, 1, 1), date(1991, 2, 1))
    partition_create = PartitionInfo.create("part", date(1991, 2, 1), date(1991, 3, 1))
    plan.existing_partitions.append(partition_existing)
    plan.partitions_to_create.append(partition_create)
    assert plan.sorted_all == [partition_existing, partition_create]
    assert plan.sorted_existing == [partition_existing]
    assert plan.sorted_to_create == [partition_create]
    assert plan.last_partition == partition_create
    assert plan.covers_date(date(1990, 12, 31)) is False
    assert plan.covers_date(date(1991, 1, 1))
    assert plan.covers_date(date(1991, 1, 31))
    assert plan.covers_date(date(1991, 2, 1))
    assert plan.covers_date(date(1991, 2, 28))
    assert plan.covers_date(date(1991, 3, 1)) is False
    assert plan.covers_date(date(1000, 1, 1)) is False
    assert plan.covers_date(date(2999, 1, 1)) is False
    plan.validate_existing_partitions()
    plan.validate_to_create()
    plan.validate_all()


def test_valid_plan_with_many():
    plan = Plan()
    partitions_existing = [
        PartitionInfo.create("part", date(1991, 1, 1), date(1991, 2, 1)),
        PartitionInfo.create("part", date(1991, 2, 1), date(1991, 3, 1)),
        PartitionInfo.create("part", date(1991, 3, 1), date(1991, 4, 1)),
        PartitionInfo.create("part", date(1991, 4, 1), date(1991, 5, 1)),
    ]
    partitions_create = [
        PartitionInfo.create("part", date(1991, 5, 1), date(1991, 6, 1)),
        PartitionInfo.create("part", date(1991, 6, 1), date(1991, 7, 1)),
        PartitionInfo.create("part", date(1991, 7, 1), date(1991, 8, 1)),
    ]
    plan.existing_partitions.extend(partitions_existing)
    plan.partitions_to_create.extend(partitions_create)
    random.shuffle(plan.existing_partitions)
    random.shuffle(plan.partitions_to_create)
    assert plan.sorted_all == partitions_existing + partitions_create
    assert plan.sorted_existing == partitions_existing
    assert plan.sorted_to_create == partitions_create
    assert plan.last_partition == partitions_create[-1]
    assert plan.covers_date(date(1990, 12, 31)) is False
    assert plan.covers_date(date(1991, 1, 1))
    assert plan.covers_date(date(1991, 1, 31))
    assert plan.covers_date(date(1991, 2, 1))
    assert plan.covers_date(date(1991, 2, 28))
    assert plan.covers_date(date(1991, 7, 31))
    assert plan.covers_date(date(1991, 8, 1)) is False
    assert plan.covers_date(date(1000, 1, 1)) is False
    assert plan.covers_date(date(2999, 1, 1)) is False
    plan.validate_existing_partitions()
    plan.validate_to_create()
    plan.validate_all()


def test_invalid_existing_partitions():
    plan = Plan(
        existing_partitions=[
            PartitionInfo.create("part", date(1991, 1, 1), date(1991, 2, 1)),
            # PartitionInfo.create("part", date(1991, 2, 1), date(1991, 3, 1)),
            PartitionInfo.create("part", date(1991, 3, 1), date(1991, 4, 1)),
        ],
        partitions_to_create=[
            PartitionInfo.create("part", date(1991, 4, 1), date(1991, 5, 1)),
            PartitionInfo.create("part", date(1991, 5, 1), date(1991, 6, 1)),
            PartitionInfo.create("part", date(1991, 6, 1), date(1991, 7, 1)),
            PartitionInfo.create("part", date(1991, 7, 1), date(1991, 8, 1)),
        ],
    )
    with pytest.raises(Exception):
        plan.validate_existing_partitions()
    plan.validate_to_create()
    with pytest.raises(Exception):
        plan.validate_all()


def test_invalid_partitions_to_create():
    plan = Plan(
        existing_partitions=[
            PartitionInfo.create("part", date(1991, 1, 1), date(1991, 2, 1)),
            PartitionInfo.create("part", date(1991, 2, 1), date(1991, 3, 1)),
            PartitionInfo.create("part", date(1991, 3, 1), date(1991, 4, 1)),
        ],
        partitions_to_create=[
            PartitionInfo.create("part", date(1991, 4, 1), date(1991, 5, 1)),
            # PartitionInfo.create("part", date(1991, 5, 1), date(1991, 6, 1)),
            PartitionInfo.create("part", date(1991, 6, 1), date(1991, 7, 1)),
            PartitionInfo.create("part", date(1991, 7, 1), date(1991, 8, 1)),
        ],
    )
    plan.validate_existing_partitions()
    with pytest.raises(Exception):
        plan.validate_to_create()
    with pytest.raises(Exception):
        plan.validate_all()


def test_invalid_partitions_to_create_missing_day():
    plan = Plan(
        existing_partitions=[
            PartitionInfo.create("part", date(1991, 1, 1), date(1991, 2, 1)),
            PartitionInfo.create("part", date(1991, 2, 2), date(1991, 3, 1)),  # missing day here
            PartitionInfo.create("part", date(1991, 3, 1), date(1991, 4, 1)),
        ],
        partitions_to_create=[
            PartitionInfo.create("part", date(1991, 4, 1), date(1991, 5, 1)),
            PartitionInfo.create("part", date(1991, 5, 1), date(1991, 6, 2)),  # missing day here
            PartitionInfo.create("part", date(1991, 6, 1), date(1991, 7, 1)),
            PartitionInfo.create("part", date(1991, 7, 1), date(1991, 8, 1)),
        ],
    )
    with pytest.raises(Exception):
        plan.validate_existing_partitions()
    with pytest.raises(Exception):
        plan.validate_to_create()
    with pytest.raises(Exception):
        plan.validate_all()



import os
from datetime import date
from typing import List
from typing import Tuple
from typing import Type

from django.db.models import Model

from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager
from django_partitioned_audit.models import SimpleAuditEntry


class SimpleAuditTableManager(AuditTableManager):
    @property
    def partitioned_table_name(self) -> str:
        return "audit_simple"

    @property
    def model(self) -> Type[Model]:
        return SimpleAuditEntry

    def create_partitioned_table(self, cursor):
        cursor.execute(
            """
        CREATE TABLE IF NOT EXISTS audit_simple (
            object_table character varying(63) NOT NULL,
            object_payload JSONB NOT NULL,
            audit_entry_created timestamp with time zone DEFAULT now() NOT NULL,
            audit_entry_created_statement_timestamp timestamp with time zone DEFAULT statement_timestamp() NOT NULL,
            audit_txid_current bigint NOT NULL,
            audit_operation character varying(1) NOT NULL
        ) PARTITION BY RANGE (audit_entry_created);
        """
        )
        cursor.execute(
            """
        CREATE INDEX IF NOT EXISTS idx_audit_simple
            ON audit_simple
            USING GIN (object_payload);
        """
        )
        cursor.execute(
            """
        CREATE OR REPLACE VIEW audit_simple_view AS (
            SELECT
                ROW_NUMBER() OVER () as "id",
                object_table,
                object_payload,
                audit_entry_created,
                audit_entry_created_statement_timestamp,
                audit_txid_current,
                audit_operation
            FROM
                audit_simple
        );
        """
        )
        cursor.execute(
            """
        CREATE OR REPLACE FUNCTION trigger_audit_entry_creator_func_v2() RETURNS TRIGGER AS $scr$
            DECLARE
                object_payload jsonb;
                trigger_operation varchar;
            BEGIN
                IF (TG_OP = 'INSERT') THEN
                    object_payload  = to_jsonb(NEW);
                    trigger_operation = 'I';
                ELSIF (TG_OP = 'UPDATE') THEN
                    object_payload  = to_jsonb(NEW);
                    trigger_operation = 'U';
                ELSIF (TG_OP = 'DELETE') THEN
                    object_payload  = to_jsonb(OLD);
                    trigger_operation = 'D';
                ELSE
                    RAISE EXCEPTION 'Unexpected TG_OP = %', TG_OP;
                END IF;

                INSERT INTO audit_simple (
                        object_table,
                        object_payload,
                        audit_txid_current,
                        audit_operation
                    )
                    SELECT
                        TG_TABLE_NAME,
                        object_payload,
                        txid_current(),
                        trigger_operation;
                RETURN NULL;
            END;
        $scr$ LANGUAGE plpgsql;
        """
        )

    def enable_audit_for_table(self, cursor, table_name: str):
        sql = """
        -- https://www.postgresql.org/message-id/1360944248465-5745434.post%40n5.nabble.com
        DO
        $src$
        BEGIN
        IF NOT EXISTS(SELECT * FROM information_schema.triggers
                      WHERE event_object_table = '{table_name}' AND trigger_name = '{trigger_name}')
            THEN

                CREATE TRIGGER {trigger_name} AFTER INSERT OR UPDATE OR DELETE ON {table_name}
                    FOR EACH ROW EXECUTE FUNCTION trigger_audit_entry_creator_func_v2();

            END IF;
        END;
        $src$
        """.format(
            table_name=table_name,  # FIXME: properly escape this to guarantee avoid SQL injection
            trigger_name="trigger_audit_entry_creator_trigger",
        )
        cursor.execute(sql)

    def disable_audit_for_table(self, cursor, table_name: str):
        sql = "DROP TRIGGER IF EXISTS {trigger_name} ON {table_name};".format(
            table_name=table_name,  # FIXME: properly escape this to guarantee avoid SQL injection
            trigger_name="trigger_audit_entry_creator_trigger",
        )
        cursor.execute(sql)

    def generate_create_partition(
        self, partition_name: str, partitioned_table: str, from_date: date, to_date: date
    ) -> Tuple[str, List[object]]:
        # TODO: find a way to escape tables names, so we don't have to use that regex
        sql_statement = f"""
        CREATE TABLE {partition_name}
            PARTITION OF {partitioned_table}
            FOR VALUES FROM (%s) TO (%s);
        """
        params = [from_date, to_date]
        return sql_statement, params

    def drop_all(self, cursor):
        assert "PYTEST_CURRENT_TEST" in os.environ

        # we need 'IF EXISTS' because data on `information_schema` is not in-sync with transaction,
        # we can get from there references to triggers, etc. that no longer exists.
        # TODO: is this the real explanation?

        for table_name, trigger_name in self.get_triggers(cursor):
            if trigger_name == "trigger_audit_entry_creator_trigger":
                cursor.execute(f"DROP TRIGGER IF EXISTS {trigger_name} ON {table_name}")

        drop_partitions = """
        DO
        $src$
        DECLARE
            item VARCHAR;
        BEGIN
            FOR item IN
                SELECT table_name FROM information_schema.tables WHERE table_name LIKE %s
            LOOP
                PERFORM 'DROP TABLE ' || item;
            END LOOP;
        END;
        $src$
        """
        params = [f"{self.partitioned_table_name}_????????_????????"]
        cursor.execute(drop_partitions, params)

        cursor.execute(f"DROP VIEW IF EXISTS {self.partitioned_table_name}_view")
        cursor.execute(f"DROP TABLE IF EXISTS {self.partitioned_table_name}")



import datetime
from typing import Iterator
from typing import List

from django.db import connection

from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager
from django_partitioned_audit.partitions.partition_info import PartitionInfo
from django_partitioned_audit.partitions.time_range_partitioning import TimeRangeGenerator


class BasePartitionManager:
    default_extra_days = 30
    postgresql_table_name_limit = 63
    partition_suffix_len = 18

    def __init__(
        self,
        time_range_generator: TimeRangeGenerator,
        extra_days=default_extra_days,
        audit_manager: AuditTableManager = None,
    ):
        self._now = datetime.datetime.now()
        self._time_range_generator = time_range_generator
        self._extra_days = extra_days
        self._audit_manager = audit_manager or AuditTableManager.get_implementation()

        assert self._extra_days >= 0
        self._validate_partitioned_table()

    @property
    def audit_manager(self) -> AuditTableManager:
        return self._audit_manager

    @property
    def _partitioned_table(self):
        return self._audit_manager.partitioned_table_name

    def _validate_partitioned_table(self):
        if len(self._partitioned_table) > (self.postgresql_table_name_limit - self.partition_suffix_len):
            raise ValueError("partitioned_table is too long, partitions will be truncated")

    @property
    def today(self) -> datetime.date:
        return self._now.date()

    def get_tables(self) -> List[str]:
        """Return the name of all the tables"""
        with connection.cursor() as cursor:
            # https://www.postgresql.org/docs/current/infoschema-tables.html
            cursor.execute(
                """
                SELECT table_name
                    FROM information_schema.tables
                    ORDER BY table_name""",
            )
            tables = [row[0] for row in cursor.fetchall()]
            return tables

    def get_existing_partitions(self, include_other: bool = False) -> Iterator[PartitionInfo]:
        """Return information about existing partitions"""
        for table_name in self.get_tables():
            partition_info = PartitionInfo.parse(table_name)
            if partition_info:
                if include_other:
                    yield partition_info
                else:
                    if partition_info.partitioned_table == self._partitioned_table:
                        yield partition_info

    def create_partitions(self, partitions: List[PartitionInfo]):
        assert partitions
        assert {pi.partitioned_table for pi in partitions} == {self._partitioned_table}

        tables = self.get_tables()
        if self._partitioned_table not in tables:
            raise ValueError(f"Partitioned table '{self._partitioned_table}' not found. Tables: {sorted(tables)}")

        with connection.cursor() as cursor:
            for part_info in partitions:
                sql, params = part_info.generate_create_partition()
                cursor.execute(sql, params)



import datetime
import os
import uuid
from datetime import timedelta
from typing import Callable
from typing import List

import pytest
from django.db import connection

from django_partitioned_audit.audit_table.advanced_audit_table_manager import AdvancedAuditTableManager
from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager
from django_partitioned_audit.audit_table.simple_audit_table_manager import SimpleAuditTableManager
from django_partitioned_audit.partitions.partition_info import PartitionInfo
from django_partitioned_audit.partitions.partition_manager_time_range import PartitionManager
from django_partitioned_audit.partitions.time_range_partitioning import DailyTimeRangeGenerator
from django_partitioned_audit.partitions.time_range_partitioning import MonthlyTimeRangeGenerator

ADVANCED_FQCN = "django_partitioned_audit.audit_table.advanced_audit_table_manager.AdvancedAuditTableManager"

skipifadvanced = pytest.mark.skipif(
    os.environ.get("DPA_AUDIT_TABLE_MANAGER") == ADVANCED_FQCN,
    reason="skip when testing AdvancedAuditTableManager",
)


@pytest.fixture
def partition_created() -> PartitionInfo:
    """
    Creates a partition for current time, so test can run and audit events inserted.
    We need this because if only the `partitioned table` is created in migration, but no partitions.
    """
    today = datetime.date.today()
    audit_tm = AuditTableManager.get_implementation()
    default_partitioned_table = audit_tm.partitioned_table_name
    info = PartitionInfo.create(default_partitioned_table, today, today + timedelta(days=2))
    sql, params = info.generate_create_partition()
    with connection.cursor() as cursor:
        cursor.execute(sql, params)

    yield info

    with connection.cursor() as cursor:
        # Why we need to delete the partition? :/
        cursor.execute(f"DROP TABLE {info.partition}")


@pytest.fixture
def partitioned_table_create():
    def func(partitioned_table):
        with connection.cursor() as cursor:
            sql = f"""
            CREATE TABLE {partitioned_table} (
                column_1 TEXT NULL,
                column_2 TEXT NULL,
                creation_timestamp timestamp with time zone DEFAULT now() NOT NULL
            ) PARTITION BY RANGE (creation_timestamp);
            """
            cursor.execute(sql)

    return func


@pytest.fixture
def very_long_named_partitioned_table(partitioned_table_create: Callable):
    """
    Creates a `partitioned table` with a name that is truncated by postgres.
    Thus, `partitions` will be truncated too.
    """
    partitioned_table = f"partitioned_table_{uuid.uuid4().hex}_long_name_lorem_ipsum_dolor_sit_amet_consectetur"
    assert len(partitioned_table) == 99, partitioned_table

    partitioned_table_create(partitioned_table)

    return partitioned_table


@pytest.fixture
def partitioned_table(partitioned_table_create: Callable):
    """
    Creates a `partitioned table` with a name that is short,
    guaranteeing that names of `partitions` won't be truncated.
    """
    partitioned_table = f"partitioned_table_{uuid.uuid4().hex[0:7]}"
    assert len(partitioned_table) == 25
    partitioned_table_create(partitioned_table)
    return partitioned_table


@pytest.fixture
def truncate_partitions():
    pm = PartitionManager(time_range_generator=MonthlyTimeRangeGenerator())
    with connection.cursor() as cursor:
        for info in pm.get_existing_partitions():
            cursor.execute(f"TRUNCATE {info.partition}")

    yield

    with connection.cursor() as cursor:
        for info in pm.get_existing_partitions():
            cursor.execute(f"DROP TABLE {info.partition}")


@pytest.fixture()
def audit_table_manager() -> AuditTableManager:
    return AuditTableManager.get_implementation()


@pytest.fixture
def settings_audit_table_manager(settings):
    def func(clazz):
        fqcn = f"{clazz.__module__}.{clazz.__qualname__}"
        settings.DPA_AUDIT_TABLE_MANAGER = fqcn

    return func


@pytest.fixture
def remove_auditing(post_test_disable_trigger):
    for am in [SimpleAuditTableManager(), AdvancedAuditTableManager()]:
        with connection.cursor() as cursor:
            am.drop_all(cursor)
        tmp = PartitionManager(audit_manager=am, time_range_generator=DailyTimeRangeGenerator())
        assert not list(tmp.get_existing_partitions())
        assert am.partitioned_table_name not in list(tmp.get_tables())

    yield


@pytest.fixture
def post_test_disable_trigger():
    yield

    # After each test the tables are truncated, but if this operation is audited after the partition
    # is deleted, the truncate() made by Django/pytest fails. Here we're doing a trick to avoid that failure.
    # Seems like this is not needed when test are successful :/ but let's do it just in case
    # TODO: this breaks encapsulation of `AuditTableManager`, maybe should be refactored
    with connection.cursor() as cursor:
        sql = """
        CREATE OR REPLACE FUNCTION if_modified_func() RETURNS TRIGGER AS $body$
        BEGIN
            RETURN NULL;
        END;
        $body$
        LANGUAGE plpgsql
        """
        cursor.execute(sql)

        sql = """
        CREATE OR REPLACE FUNCTION trigger_audit_entry_creator_func_v2() RETURNS TRIGGER AS $body$
        BEGIN
            RETURN NULL;
        END;
        $body$
        LANGUAGE plpgsql
        """
        cursor.execute(sql)


@pytest.fixture
def use_audit_table_manager(settings, remove_auditing):
    def func(audit_table_manager_class, enable_models: List):
        clazz = f"{audit_table_manager_class.__module__}.{audit_table_manager_class.__qualname__}"
        settings.DPA_AUDIT_TABLE_MANAGER = clazz

        audit_manager = AuditTableManager.get_implementation()
        assert isinstance(audit_manager, audit_table_manager_class)
        with connection.cursor() as cursor:
            audit_manager.create_partitioned_table(cursor)
            for a_model in enable_models:
                audit_manager.enable_audit_for_table(cursor, a_model._meta.db_table)

        today = datetime.date.today()
        audit_tm = AuditTableManager.get_implementation()
        default_partitioned_table = audit_tm.partitioned_table_name
        info = PartitionInfo.create(default_partitioned_table, today, today + timedelta(days=2))
        sql, params = info.generate_create_partition()
        with connection.cursor() as cursor:
            cursor.execute(sql, params)

    return func


# def _table_exists(table_name: str) -> bool:
#     with transaction.atomic():
#         with connection.cursor() as cursor:
#             try:
#                 cursor.execute(f"select * from {table_name} limit 1")
#             except ProgrammingError as err:
#                 if re.search(rf"relation.*{table_name}.*does not exist", err.args[0].splitlines()[0]):
#                     return False
#                 else:
#                     raise
#             return True



import uuid

from django.db import models


class Customer(models.Model):
    name = models.CharField(max_length=100)
    active = models.BooleanField(default=False)


class Product(models.Model):
    product_id = models.BigAutoField(primary_key=True)
    name = models.CharField(max_length=100)


class Invoice(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=100)


class InvoiceItem(models.Model):
    name = models.CharField(max_length=100)



import json
import uuid

import requests
from app.models import Customer
from django.core.management.base import BaseCommand
from django.urls import reverse

from django_partitioned_audit.models import SimpleAuditEntry


class Command(BaseCommand):
    help = "Closes the specified poll for voting"

    def add_arguments(self, parser):
        parser.add_argument("--server-url", default="http://127.0.0.1:8000")

    def handle(self, *args, **options):
        server_url = options["server_url"]

        print(f"server_url={server_url}")
        original_name = str(uuid.uuid4())
        new_name = str(uuid.uuid4())
        customer = Customer.objects.create(name=original_name)
        url = f"{server_url}{reverse('customer/update', args=[customer.pk])}"
        response = requests.post(url, data=dict(name=new_name))
        response.raise_for_status()

        print("Finished. Now checking...")

        # Assert model was created and audit exists
        for audit in SimpleAuditEntry.objects.filter(object_table="app_customer"):
            # FIXME: filter by `id`
            # Before was possible just by `filter(object_pk=customer.pk)`
            print("-" * 120)
            print(f"Audit: {audit}")
            print(f"object_payload: {json.dumps(json.loads(audit.object_payload), indent=4)}")



import uuid

import pytest
from app.models import Customer
from app.models import Invoice
from app.models import Product

from django_partitioned_audit.audit_table.advanced_audit_table_manager import AdvancedAuditTableManager
from django_partitioned_audit.models import AdvancedAuditEntry
from django_partitioned_audit.models import Operation
from django_partitioned_audit.models import SimpleAuditEntry
from tests.conftest import skipifadvanced

# pylint: skip-file


@skipifadvanced
@pytest.mark.django_db(transaction=True)
def test_filter_model(partition_created):
    initial_name = str(uuid.uuid4())
    new_name = str(uuid.uuid4())

    assert SimpleAuditEntry.objects.filter_model(Invoice).count() == 0
    assert SimpleAuditEntry.objects.filter_model(Product).count() == 0
    assert SimpleAuditEntry.objects.filter_model(Customer).count() == 0

    Invoice.objects.create(name=str(uuid.uuid4()))
    Product.objects.create(name=str(uuid.uuid4()))
    Product.objects.create(name=str(uuid.uuid4()))
    Customer.objects.create(name=str(uuid.uuid4()))
    Customer.objects.create(name=str(uuid.uuid4()))
    customer = Customer.objects.create(name=initial_name)

    assert SimpleAuditEntry.objects.filter_model(Invoice).count() == 1
    assert SimpleAuditEntry.objects.filter_model(Product).count() == 2
    assert SimpleAuditEntry.objects.filter_model(Customer).count() == 3

    customer.name = new_name
    customer.save()

    assert SimpleAuditEntry.objects.filter_model(Customer).count() == 4


@skipifadvanced
@pytest.mark.django_db(transaction=True)
def test_filter_payload(partition_created):
    Customer.objects.create(name=str(uuid.uuid4()))
    Customer.objects.create(name=str(uuid.uuid4()))

    initial_name = str(uuid.uuid4())
    new_name = str(uuid.uuid4())
    customer = Customer.objects.create(name=initial_name)

    assert SimpleAuditEntry.objects.filter_payload(name=initial_name).count() == 1
    assert SimpleAuditEntry.objects.filter_payload(name=new_name).count() == 0

    customer.active = False
    customer.save()

    assert SimpleAuditEntry.objects.filter_payload(name=initial_name).count() == 2
    assert SimpleAuditEntry.objects.filter_payload(name=new_name).count() == 0

    customer.name = new_name
    customer.save()

    assert SimpleAuditEntry.objects.filter_payload(name=initial_name).count() == 2
    assert SimpleAuditEntry.objects.filter_payload(name=new_name).count() == 1


@skipifadvanced
@pytest.mark.django_db(transaction=True)
def test_filter_by_primary_key(partition_created):
    Invoice.objects.create(name=str(uuid.uuid4()))
    Product.objects.create(name=str(uuid.uuid4()))
    Customer.objects.create(name=str(uuid.uuid4()))

    invoice = Invoice.objects.create(name=str(uuid.uuid4()))
    product = Product.objects.create(name=str(uuid.uuid4()))
    customer = Customer.objects.create(name=str(uuid.uuid4()))

    audit_entry = SimpleAuditEntry.objects.filter_model(Invoice).filter_payload(id=str(invoice.pk)).get()
    assert audit_entry.get_row_data()["name"] == invoice.name

    audit_entry = SimpleAuditEntry.objects.filter_model(Product).filter_payload(product_id=int(product.pk)).get()
    assert audit_entry.get_row_data()["name"] == product.name

    audit_entry = SimpleAuditEntry.objects.filter_model(Customer).filter_payload(id=int(customer.pk)).get()
    assert audit_entry.get_row_data()["name"] == customer.name


@skipifadvanced
@pytest.mark.django_db(transaction=True)
def test_filter_by_primary_key_advanced(use_audit_table_manager):
    use_audit_table_manager(AdvancedAuditTableManager, [Invoice, Product, Customer])

    Invoice.objects.create(name=str(uuid.uuid4()))
    Product.objects.create(name=str(uuid.uuid4()))
    Customer.objects.create(name=str(uuid.uuid4()))

    invoice = Invoice.objects.create(name=str(uuid.uuid4()))
    product = Product.objects.create(name=str(uuid.uuid4()))
    customer = Customer.objects.create(name=str(uuid.uuid4()))

    audit_entry = AdvancedAuditEntry.objects.filter_model(Invoice).filter_payload(id=str(invoice.pk)).get()
    assert audit_entry.get_row_data()["name"] == invoice.name

    audit_entry = AdvancedAuditEntry.objects.filter_model(Product).filter_payload(product_id=str(product.pk)).get()
    assert audit_entry.get_row_data()["name"] == product.name
    # Pay attention at `filter_payload()`... For `AdvancedAuditEntry` we use `str(product.pk)`,
    # because even tough JSONB supports non-str, in the table we're using HSTORE

    audit_entry = AdvancedAuditEntry.objects.filter_model(Customer).filter_payload(id=str(customer.pk)).get()
    assert audit_entry.get_row_data()["name"] == customer.name
    # The same here with `filter_payload(id=str(customer.pk))`


@skipifadvanced
@pytest.mark.django_db(transaction=True)
def test_filter_operation(partition_created):
    assert SimpleAuditEntry.objects.filter_model(Invoice).count() == 0
    assert SimpleAuditEntry.objects.filter_model(Product).count() == 0
    assert SimpleAuditEntry.objects.filter_model(Customer).count() == 0

    Invoice.objects.create(name=str(uuid.uuid4()))
    product = Product.objects.create(name=str(uuid.uuid4()))
    customer = Customer.objects.create(name=str(uuid.uuid4()))

    customer.name = str(uuid.uuid4())
    customer.save()

    product.delete()

    assert SimpleAuditEntry.objects.filter_model(Invoice).count() == 1
    assert SimpleAuditEntry.objects.filter_model(Product).count() == 2
    assert SimpleAuditEntry.objects.filter_model(Customer).count() == 2

    assert SimpleAuditEntry.objects.filter_operation(Operation.INSERT).count() == 3
    assert SimpleAuditEntry.objects.filter_operation(Operation.UPDATE).count() == 1
    assert SimpleAuditEntry.objects.filter_operation(Operation.DELETE).count() == 1

    tables = SimpleAuditEntry.objects.filter_operation(Operation.INSERT).values_list("object_table", flat=True)
    assert set(tables) == {"app_customer", "app_invoice", "app_product"}

    tables = SimpleAuditEntry.objects.filter_operation(Operation.UPDATE).values_list("object_table", flat=True)
    assert set(tables) == {"app_customer"}

    tables = SimpleAuditEntry.objects.filter_operation(Operation.DELETE).values_list("object_table", flat=True)
    assert set(tables) == {"app_product"}



from datetime import date

import pytest

from django_partitioned_audit.partitions.partition_info import PartitionInfo


def test_partition_info_parse():
    partition_info = PartitionInfo.parse("some_partitioned_table_19980102_19990304")
    assert partition_info
    assert partition_info.partition == "some_partitioned_table_19980102_19990304"
    assert partition_info.partitioned_table == "some_partitioned_table"
    assert partition_info.from_date == date(1998, 1, 2)
    assert partition_info.to_date == date(1999, 3, 4)


@pytest.mark.parametrize(
    "table_name",
    [
        "some_partitioned_table_19980102_19970304",  # year after
        "some_partitioned_table_19980402_19980304",  # month after
        "some_partitioned_table_19990305_19990304",  # day after
        "some_partitioned_table_19990304_19990304",  # same day
    ],
)
def test_partition_info_parse_fails(table_name):
    with pytest.raises(ValueError):
        PartitionInfo.parse(table_name)


def test_partition_info_create():
    partition_info = PartitionInfo.create("foo", date(1998, 1, 2), date(1999, 3, 4))
    assert partition_info.partition == "foo_19980102_19990304"
    assert partition_info.from_date == date(1998, 1, 2)
    assert partition_info.to_date == date(1999, 3, 4)


def test_generate_create_partition():
    partition_info = PartitionInfo.create("foo", date(1998, 1, 2), date(1999, 3, 4))
    sql, params = partition_info.generate_create_partition()
    sql_line = " ".join([_.strip() for _ in sql.splitlines() if _.strip()])
    sql_line = " ".join([_.strip() for _ in sql_line.split() if _.strip()])
    sql_line = sql_line.lower()
    sql_line = sql_line.replace("%s", "{}")
    # not exactly the same interpolation done by db driver, but usable for tests
    final_sql = sql_line.format(*params)
    expected_sql = (
        """create table foo_19980102_19990304 """
        """partition of foo """
        """for values from (1998-01-02) to (1999-03-04);"""
    )
    assert final_sql == expected_sql



import logging

from django.apps import AppConfig
from django.db import connections
from django.db.models.signals import post_migrate
from django.dispatch import receiver

from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager
from django_partitioned_audit.utils import get_disabled_models
from django_partitioned_audit.utils import get_enabled_models

logger = logging.getLogger(__name__)


@receiver(post_migrate)
def audit_run_post_migrate(app_config: AppConfig, verbosity, using, **kwargs):  # pylint: disable=unused-argument
    enabled_models = set(get_enabled_models(app_config))
    disabled_models = set(get_disabled_models(app_config))

    if disabled_models & enabled_models:
        print(f"Model are at the same time 'enabled' and 'disabled': {disabled_models & enabled_models}")

    disabled_models = disabled_models - enabled_models

    with connections[using].cursor() as cursor:
        # For now, we use the 'using' connection we received. Not sure if that's the best approach.
        audit_table_manager: AuditTableManager = AuditTableManager.get_implementation()

        # Let's (re)create the partitioned table and other objects
        audit_table_manager.create_partitioned_table(cursor)

        # Let's create the triggers
        for model_class_name in enabled_models:
            model_class = app_config.get_model(model_class_name)
            table_name = model_class._meta.db_table  # pylint: disable=protected-access
            if verbosity >= 1:
                print(f" - Enable auditing for '{model_class}' (table '{table_name}')")
            audit_table_manager.enable_audit_for_table(cursor, table_name)

        # and remove the ones that needs to be removed
        for model_class_name in disabled_models:
            model_class = app_config.get_model(model_class_name)
            table_name = model_class._meta.db_table  # pylint: disable=protected-access
            if verbosity >= 1:
                print(f" - Disabling auditing for '{model_class}' (table '{table_name}')")
            audit_table_manager.disable_audit_for_table(cursor, table_name)


# *****************************************************************************
# About original implementation of signal handler
# *****************************************************************************
# ENABLE_TRIGGER_SQL = """
# -- This looks dangerous, but seems to be the recommended approach.
# -- Nevertheless, it looks dangerous, and if for some reason a CASCADE DROP
# --  happens (it should't, but anyway), so, better try some other alternative
# BEGIN;
# DROP TRIGGER IF EXISTS trigger_audit_entry_creator_trigger ON {table_name};
# CREATE TRIGGER trigger_audit_entry_creator_trigger
#     AFTER INSERT OR UPDATE OR DELETE ON {table_name}
#     FOR EACH ROW EXECUTE FUNCTION trigger_audit_entry_creator_func();
#
# COMMIT;
# """
# *****************************************************************************



import dataclasses
from dataclasses import field
from datetime import date
from datetime import timedelta
from enum import Enum
from typing import Iterator
from typing import List
from typing import Optional

from django_partitioned_audit.partitions.partition_info import PartitionInfo


class Status(Enum):
    OK = 1
    MISSING_NEXT = 2
    MISSING_CURRENT = 3


@dataclasses.dataclass
class Plan:
    existing_partitions: List[PartitionInfo] = field(default_factory=list)
    partitions_to_create: List[PartitionInfo] = field(default_factory=list)
    status: Optional[Status] = None

    @property
    def sorted_existing(self) -> List[PartitionInfo]:
        return sorted(self.existing_partitions, key=lambda _: _.from_date)

    @property
    def sorted_to_create(self) -> List[PartitionInfo]:
        return sorted(self.partitions_to_create, key=lambda _: _.from_date)

    @property
    def sorted_all(self) -> List[PartitionInfo]:
        return sorted(self.partitions_to_create + self.existing_partitions, key=lambda _: _.from_date)

    @property
    def last_partition(self) -> Optional[PartitionInfo]:
        try:
            return self.sorted_all[-1]
        except IndexError:
            return None

    def covers_date(self, day_to_cover: date) -> bool:
        return self.get_partition_for_date(self.sorted_all, day_to_cover) is not None

    @classmethod
    def get_partition_for_date(cls, partitions: List[PartitionInfo], reference_day: date) -> Optional[PartitionInfo]:
        for part_info in partitions:
            if part_info.from_date <= reference_day < part_info.to_date:
                return part_info
        return None

    @classmethod
    def _validate(cls, partitions: Iterator[PartitionInfo]):
        # TODO: when empty, should we fail?
        last_to_date = None
        for pi in partitions:
            if last_to_date is None:
                last_to_date = pi.to_date
            else:
                if last_to_date != pi.from_date:
                    raise Exception(f"Invalid: last_to_date={last_to_date} pi={pi}")
                last_to_date = pi.to_date

    def validate_existing_partitions(self):
        self._validate(self.sorted_existing)

    def validate_to_create(self):
        self._validate(self.sorted_to_create)

    def validate_all(self):
        self._validate(self.sorted_all)

    def calculate_status(self, today: date, extra_days: int):
        assert self.status is None
        assert extra_days >= 0
        assert self.existing_partitions or self.partitions_to_create  # one or both must contain something
        if self.existing_partitions and not self.partitions_to_create:
            # no new partitions, this means all is good, current time range and extra days are covered
            self.status = Status.OK
            # Just in case, check today is covered by existing
            assert self.get_partition_for_date(self.sorted_existing, today) is not None
            # Juts in case, check extra days
            assert self.sorted_all[-1].to_date >= today + timedelta(days=extra_days)
            return

        if not self.existing_partitions and self.partitions_to_create:
            # no existing partitions, this must be the first time the command is run
            self.status = Status.MISSING_CURRENT
            # Just in case, check today is covered by 'to create'
            assert self.get_partition_for_date(self.sorted_to_create, today) is not None
            # Juts in case, check extra days
            assert self.sorted_all[-1].to_date >= today + timedelta(days=extra_days)
            return

        if self.existing_partitions and self.partitions_to_create:
            # both existing partitions and to create... let's see is we need to create for current time range
            current_time_range_in_existing = self.get_partition_for_date(self.sorted_existing, today) is not None
            current_time_range_in_to_create = self.get_partition_for_date(self.sorted_to_create, today) is not None
            assert {current_time_range_in_existing, current_time_range_in_to_create} == {True, False}

            if current_time_range_in_to_create:
                # This is the worst situation. There are existing tables, this is not the first time the system
                # is run, but there was no partition for current time. If the partitioned table was in use, the
                # system might have lacked the required partitions to be able to insert data.
                self.status = Status.MISSING_CURRENT
                # Just in case, check today is covered by 'to create'
                assert self.get_partition_for_date(self.sorted_to_create, today) is not None
            else:
                self.status = Status.MISSING_NEXT
                # Just in case, check today is covered by existing
                assert self.get_partition_for_date(self.sorted_existing, today) is not None

            # Juts in case, check extra days
            assert self.sorted_all[-1].to_date >= today + timedelta(days=extra_days)



from app.models import Customer
from app.models import Invoice
from app.models import Product
from django.contrib import admin

from django_partitioned_audit.admin import TriggerAuditEntryAdmin
from django_partitioned_audit.models import SimpleAuditEntry


@admin.register(Customer)
class CustomerAdmin(admin.ModelAdmin):
    pass


@admin.register(Invoice)
class InvoiceAdmin(admin.ModelAdmin):
    pass


@admin.register(Product)
class ProductAdmin(admin.ModelAdmin):
    pass


@admin.register(SimpleAuditEntry)
class CustomTriggerAuditEntryAdmin(TriggerAuditEntryAdmin):
    pass



import uuid

import pytest
import requests
from app.models import Customer
from app.models import Invoice
from app.models import Product
from django.urls import reverse

# pylint: skip-file


"""
These tests exist to make sure the application we use to test, actually works.
These tests do not test `django-partitioned-audit` functionality.
"""


@pytest.mark.django_db(transaction=True)
def test_create_customer_works_with_audit_disabled(live_server, remove_auditing):
    name = str(uuid.uuid4())
    response = requests.post(f"{live_server}{reverse('customer/create')}", data=dict(name=name))
    response.raise_for_status()
    assert Customer.objects.filter(name=name).exists()


@pytest.mark.django_db(transaction=True)
def test_update_customer_works_with_audit_disabled(live_server, remove_auditing):
    orig_name = str(uuid.uuid4())
    new_name = str(uuid.uuid4())
    customer = Customer.objects.create(name=orig_name)
    assert Customer.objects.filter(name=orig_name).exists()

    request_url = f"{live_server}{reverse('customer/update', args=[customer.pk])}"
    response = requests.post(request_url, data=dict(name=new_name))
    response.raise_for_status()
    assert not Customer.objects.filter(name=orig_name).exists()
    assert Customer.objects.filter(name=new_name).exists()


@pytest.mark.django_db(transaction=True)
def test_create_invoice_works_with_audit_disabled(live_server, remove_auditing):
    name = str(uuid.uuid4())
    response = requests.post(f"{live_server}{reverse('invoice/create')}", data=dict(name=name))
    response.raise_for_status()
    assert Invoice.objects.filter(name=name).exists()


@pytest.mark.django_db(transaction=True)
def test_create_product_works_with_audit_disabled(live_server, remove_auditing):
    name = str(uuid.uuid4())
    response = requests.post(f"{live_server}{reverse('product/create')}", data=dict(name=name))
    response.raise_for_status()
    assert Product.objects.filter(name=name).exists()



# Generated by Django 3.0.11 on 2020-12-20 14:17

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Customer",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=100)),
                ("active", models.BooleanField(default=False)),
            ],
        ),
    ]



from datetime import date
from typing import List
from typing import Tuple
from typing import Type

from django.db.models import Model

from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager


class DummyAuditTableManager(AuditTableManager):
    def __init__(self, partitioned_table_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.custom_partitioned_table_name = partitioned_table_name

    @property
    def partitioned_table_name(self) -> str:
        return self.custom_partitioned_table_name

    @property
    def model(self) -> Type[Model]:
        pass

    def create_partitioned_table(self, cursor):
        pass

    def enable_audit_for_table(self, cursor, table_name: str):
        pass

    def disable_audit_for_table(self, cursor, table_name: str):
        pass

    def drop_all(self, cursor):
        pass

    def generate_create_partition(
        self, partition_name: str, partitioned_table: str, from_date: date, to_date: date
    ) -> Tuple[str, List[object]]:
        pass



import itertools
import os
import uuid
from concurrent import futures
from concurrent.futures.thread import ThreadPoolExecutor
from typing import List

import pytest
import requests
from app.models import Customer
from django.urls import reverse

from django_partitioned_audit.models import SimpleAuditEntry
from tests.conftest import skipifadvanced

# pylint: skip-file


@skipifadvanced
@pytest.mark.skipif(os.environ.get("SKIP_SLOW"), reason="skipping slow tests")
@pytest.mark.django_db(transaction=True)
def test_concurrent_modifications(live_server, partition_created):
    customer = Customer.objects.create(name=str(uuid.uuid4()))
    urls = [
        f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
        f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=100&post-wait-ms=0",
        f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=100",
        f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=100&post-wait-ms=100",
    ]

    def _update_customer(url: str, new_name: str):
        response = requests.post(url, data=dict(name=new_name))
        response.raise_for_status()

    new_names = [str(uuid.uuid4()) for _ in range(50)]

    # We can use a with statement to ensure threads are cleaned up promptly
    with ThreadPoolExecutor(max_workers=20) as executor:
        # Start the load operations and mark each future with its URL
        url_iterator = itertools.cycle(urls)
        future_to_url = {
            executor.submit(_update_customer, next(url_iterator), new_name): new_name for new_name in new_names
        }
        for future in futures.as_completed(future_to_url):
            # new_name = future_to_url[future]
            try:
                future.result()
            except Exception:
                raise
            else:
                print("+", end="")

    # We need one audit entry for creation, then 1 audit entry for each update
    assert SimpleAuditEntry.objects.all().count() == len(new_names) + 1
    audit_entries: List[SimpleAuditEntry] = SimpleAuditEntry.objects.all()
    names_in_audit = [_.get_row_data()["name"] for _ in audit_entries]
    assert set(names_in_audit) == set([customer.name] + new_names)



import uuid

import pytest
import requests
from app.models import Customer
from django.urls import reverse

from django_partitioned_audit.audit_table.advanced_audit_table_manager import AdvancedAuditTableManager
from django_partitioned_audit.audit_table.simple_audit_table_manager import SimpleAuditTableManager
from django_partitioned_audit.models import AdvancedAuditEntry
from django_partitioned_audit.models import Operation
from django_partitioned_audit.models import SimpleAuditEntry

# pylint: skip-file


@pytest.mark.parametrize(
    "audit_table_manager_class,model_class",
    [[SimpleAuditTableManager, SimpleAuditEntry], [AdvancedAuditTableManager, AdvancedAuditEntry]],
)
def test_create(audit_table_manager_class, model_class, live_server, use_audit_table_manager):
    use_audit_table_manager(audit_table_manager_class, [Customer])
    assert model_class.objects.all().count() == 0

    name = str(uuid.uuid4())
    data = dict(name=name)
    response = requests.post(f"{live_server}{reverse('customer/create')}", data=data)
    response.raise_for_status()
    assert Customer.objects.filter(name=name).exists()

    # Assert model was created and audit exists
    audit_entry = model_class.objects.all().get()
    assert audit_entry.is_insert()
    assert audit_entry.get_row_data()["name"] == name


@pytest.mark.parametrize(
    "audit_table_manager_class,model_class",
    [[SimpleAuditTableManager, SimpleAuditEntry], [AdvancedAuditTableManager, AdvancedAuditEntry]],
)
def test_update(audit_table_manager_class, model_class, live_server, use_audit_table_manager):
    use_audit_table_manager(audit_table_manager_class, [Customer])

    original_name = str(uuid.uuid4())
    customer = Customer.objects.create(name=original_name)

    new_name = str(uuid.uuid4())
    data = dict(name=new_name)
    response = requests.post(f"{live_server}{reverse('customer/update', args=[customer.pk])}", data=data)
    response.raise_for_status()
    assert Customer.objects.filter(name=new_name).exists()

    # Assert model was created and audit exists
    audit_entry = model_class.objects.filter_operation(Operation.UPDATE).get()
    assert audit_entry.is_update()
    assert audit_entry.get_row_data()["name"] == new_name


@pytest.mark.parametrize(
    "audit_table_manager_class,model_class",
    [[SimpleAuditTableManager, SimpleAuditEntry], [AdvancedAuditTableManager, AdvancedAuditEntry]],
)
def test_delete(audit_table_manager_class, model_class, live_server, use_audit_table_manager):
    use_audit_table_manager(audit_table_manager_class, [Customer])

    name = str(uuid.uuid4())
    customer = Customer.objects.create(name=name)

    response = requests.post(f"{live_server}{reverse('customer/delete', args=[customer.pk])}")
    response.raise_for_status()
    assert not Customer.objects.filter(name=name).exists()

    # Assert model was created and audit exists
    audit_entry = model_class.objects.filter_operation(Operation.DELETE).get()
    assert audit_entry.is_delete()
    assert audit_entry.get_row_data()["name"] == name



import psycopg2.extensions

DATABASES = {
    "default": {
        "NAME": "django_partitioned_audit_app",
        "ENGINE": "django.db.backends.postgresql",
        "USER": "test12",
        "PASSWORD": "test12",
        "HOST": "localhost",
        "PORT": 55012,
        "OPTIONS": {
            # THIS IS THE DEFAULT, BUT BETTER BE SPECIFIC
            "isolation_level": psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED,
        },
        "ATOMIC_REQUESTS": True,
    },
}



import os

from .settings_boilerplate import *  # noqa

# pylint: skip-file

DEBUG = True

if "DPA_AUDIT_TABLE_MANAGER" in os.environ:
    DPA_AUDIT_TABLE_MANAGER = os.environ["DPA_AUDIT_TABLE_MANAGER"]

INSTALLED_APPS = [
    # Django applications
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.postgres",
    # django partitioned audit & test applications
    "django_partitioned_audit",
    "app.apps.AppConfig",
    "second_app.apps.SecondaryAppConfig",
]

# By default, let's import some db configuration, to make sure it will work
# when using this file directly

from .settings_postgres15 import *  # noqa



import dataclasses
import sys
from typing import List

from django.core.management.base import BaseCommand

from django_partitioned_audit.partitions.partition_info import PartitionInfo
from django_partitioned_audit.partitions.partition_manager_plan import Plan
from django_partitioned_audit.partitions.partition_manager_plan import Status
from django_partitioned_audit.partitions.partition_manager_time_range import PartitionManager
from django_partitioned_audit.partitions.time_range_partitioning import DailyTimeRangeGenerator
from django_partitioned_audit.partitions.time_range_partitioning import MonthlyTimeRangeGenerator
from django_partitioned_audit.partitions.time_range_partitioning import WeeklyTimeRangeGenerator

ACTION_NAMES = ["list", "list-all", "simulate", "create"]

TIME_RANGE_GENERATORS = {
    "MonthlyTimeRangeGenerator": MonthlyTimeRangeGenerator,
    "WeeklyTimeRangeGenerator": WeeklyTimeRangeGenerator,
    "DailyTimeRangeGenerator": DailyTimeRangeGenerator,
}


@dataclasses.dataclass
class PrettyTableDummy:
    rows: List[List[str]] = dataclasses.field(default_factory=list)

    def add_row(self, row: List[str]):
        self.rows.append(row)

    def __str__(self):
        lines = [" | ".join([str(_) for _ in cols]) for cols in self.rows]
        return "\n".join(sorted(lines))


try:
    from prettytable import PrettyTable
except ImportError:
    PrettyTable = PrettyTableDummy


class Command(BaseCommand):
    help = "Makes sure all the required partition tables exists"

    def add_arguments(self, parser):
        parser.add_argument("actions", nargs="+", choices=ACTION_NAMES)
        parser.add_argument("--extra-days", type=int, default=PartitionManager.default_extra_days)
        parser.add_argument(
            "--time-range-generator",
            type=str,
            default="MonthlyTimeRangeGenerator",
            choices=list(TIME_RANGE_GENERATORS.keys()),
        )

    def handle(self, *args, **options):
        time_range_generator_class = TIME_RANGE_GENERATORS[options["time_range_generator"]]
        pm = PartitionManager(
            extra_days=options["extra_days"],
            time_range_generator=time_range_generator_class(),
        )

        for action in options["actions"]:
            if action == "list":
                self._run_list(pm, **options)
            elif action == "list-all":
                self._run_list(pm, list_all=True, **options)
            elif action == "simulate":
                self._run_simulate(pm, **options)
            elif action == "create":
                self._run_create(pm, **options)
            else:
                raise NotImplementedError(f"Action '{action}' not implemented")

    def _run_list(self, pm: PartitionManager, list_all=False, **options):
        pretty = PrettyTable()
        pretty.field_names = ["table_name", "from_date", "to_date"]
        pretty.padding_width = 2
        pretty.sortby = "table_name"
        pretty.align = "l"

        for table_name in pm.get_tables():
            pi = PartitionInfo.parse(table_name)
            if pi:
                pretty.add_row([pi.partition, pi.from_date, pi.to_date])
            else:
                if list_all:
                    pretty.add_row([table_name, "", ""])

        self.stdout.write(str(pretty))

    def _run_simulate(self, pm: PartitionManager, exit_with_status=False, **options) -> Plan:
        pretty = PrettyTable()
        pretty.field_names = ["table_name", "from_date", "to_date", "status"]
        pretty.padding_width = 2
        pretty.sortby = "table_name"
        pretty.align = "l"
        simulation = pm.generate_plan()

        status_prefix = "⚠" if simulation.status == Status.MISSING_CURRENT else "❌"
        for pi in simulation.existing_partitions:
            pretty.add_row([pi.partition, pi.from_date, pi.to_date, "✓ exists"])
        for pi in simulation.partitions_to_create:
            pretty.add_row([pi.partition, pi.from_date, pi.to_date, f"{status_prefix} to create"])

        self.stdout.write(str(pretty))

        if exit_with_status:
            if simulation.status == Status.OK:
                sys.exit(0)
            elif simulation.status == Status.MISSING_CURRENT:
                print("ERROR: required partition to store data for current period do not exist", file=self.stdout)
                sys.exit(1)
            elif simulation.status == Status.MISSING_NEXT:
                sys.exit(0)
            else:
                raise NotImplementedError(f"Unexpected status: {simulation.status}")

        return simulation

    def _run_create(self, pm: PartitionManager, **options):
        plan = self._run_simulate(pm, exit_with_status=False)

        for pi in plan.partitions_to_create:
            sql, params = pi.generate_create_partition()
            sql_line = " ".join([_.strip() for _ in sql.splitlines() if _.strip()])
            self.stdout.write(f"sql: {sql_line} / params: {params}\n")

        if plan.partitions_to_create:
            pm.create_partitions(plan.partitions_to_create)
        else:
            self.stdout.write("All required partitions exists :D\n")



from django.contrib import admin


class TriggerAuditEntryAdmin(admin.ModelAdmin):
    list_display = ("audit_entry_created", "object_table", "audit_operation")
    list_filter = ("object_table", "audit_operation")

    def save_form(self, request, form, change):
        raise NotImplementedError()

    def save_formset(self, request, form, formset, change):
        raise NotImplementedError()

    def save_model(self, request, obj, form, change):
        raise NotImplementedError()

    def save_related(self, request, form, formsets, change):
        raise NotImplementedError()

    def delete_view(self, request, object_id, extra_context=None):
        raise NotImplementedError()

    def delete_model(self, request, obj):
        raise NotImplementedError()

    def delete_queryset(self, request, queryset):
        raise NotImplementedError()

    # FIXME: add other methods to make Admin RO
    # TODO: maybe this is not required if we change permissions in DB
    # TODO: probably multi-db hints might be a more secure and transparent way
    # https://docs.djangoproject.com/en/3.0/topics/db/multi-db/#topics-db-multi-db-hints



import uuid

import pytest
import requests
from app.models import Customer
from django.db import OperationalError
from django.urls import reverse

from django_partitioned_audit.audit_table.advanced_audit_table_manager import AdvancedAuditTableManager
from django_partitioned_audit.audit_table.simple_audit_table_manager import SimpleAuditTableManager
from django_partitioned_audit.models import AdvancedAuditEntry
from django_partitioned_audit.models import SimpleAuditEntry

# pylint: skip-file


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize(
    "audit_table_manager_class,model_class",
    [[SimpleAuditTableManager, SimpleAuditEntry], [AdvancedAuditTableManager, AdvancedAuditEntry]],
)
def test_delete_audit_entries_fail(audit_table_manager_class, model_class, live_server, use_audit_table_manager):
    use_audit_table_manager(audit_table_manager_class, [Customer])
    name = str(uuid.uuid4())
    data = dict(name=name)
    response = requests.post(f"{live_server}{reverse('customer/create')}", data=data)
    response.raise_for_status()
    assert Customer.objects.filter(name=name).exists()

    # Assert model cannot be deleted
    audit_entry = model_class.objects.all().get()
    with pytest.raises(OperationalError, match=r"cannot delete from view"):
        audit_entry.delete()



from datetime import date
from typing import Callable
from unittest.mock import Mock
from unittest.mock import patch
from uuid import uuid4

import pytest
from django.db import ProgrammingError

from django_partitioned_audit.partitions.partition_info import PartitionInfo
from django_partitioned_audit.partitions.partition_manager_time_range import PartitionManager
from django_partitioned_audit.partitions.time_range_partitioning import MonthlyTimeRangeGenerator as MTRG
from tests.dummies import DummyAuditTableManager

# pylint: skip-file


@pytest.mark.django_db(transaction=True)
def test_create_partition(partitioned_table: str):
    pm = PartitionManager(audit_manager=DummyAuditTableManager(partitioned_table), time_range_generator=MTRG())
    assert not list(pm.get_existing_partitions())

    pm.create_partitions(
        [
            PartitionInfo.create(
                partitioned_table=partitioned_table,
                from_date=date(1999, 1, 1),
                to_date=date(1999, 2, 1),
            )
        ]
    )

    assert set(pm.get_existing_partitions()) == {PartitionInfo.parse(f"{partitioned_table}_19990101_19990201")}


@pytest.mark.django_db(transaction=True)
def test_create_second_long_partition_fails(partitioned_table_create: Callable):
    # Creates a `partitioned table` with a name that is not truncated by postgres, but is long enough
    # for partitions to be truncated when adding the suffix.
    # "abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcdefghi_abc"  bad
    # "abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcdefgh_1999"  bad
    # "abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcd_19990101"  bad
    # "abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcdefghi_19990101_1999"  bad
    # "abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcde_19990101_19990101"  good

    bad = f"abcdefghi_{uuid4().hex[0:9]}_{uuid4().hex[0:9]}_{uuid4().hex[0:9]}_{uuid4().hex[0:9]}_abcdefgh"
    partition_1 = PartitionInfo.create(bad, date(1999, 1, 1), date(1999, 2, 1))
    partition_2 = PartitionInfo.create(bad, date(1999, 2, 1), date(1999, 3, 1))

    partitioned_table_create(bad)

    with patch.object(PartitionManager, "_validate_partitioned_table", new_callable=Mock):
        pm = PartitionManager(audit_manager=DummyAuditTableManager(bad), time_range_generator=MTRG())
        pm.create_partitions([partition_1])
        with pytest.raises(ProgrammingError, match=r'relation.*abcdefghi_[a-f0-9_]+_abcdefgh_1999" already exists'):
            pm.create_partitions([partition_2])



import psycopg2.extensions

# pylint: skip-file

INSTALLED_APPS = [
    "django.contrib.postgres",
    "django.contrib.contenttypes",
    "django_partitioned_audit",
]

DATABASES = {
    "default": {
        "NAME": "sdist_test_install",
        "ENGINE": "django.db.backends.postgresql",
        "USER": "test15",
        "PASSWORD": "test15",
        "HOST": "localhost",
        "PORT": 55015,
        "OPTIONS": {
            # THIS IS THE DEFAULT, BUT BETTER BE SPECIFIC
            "isolation_level": psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED,
        },
        "ATOMIC_REQUESTS": True,
    },
}



import os  # noqa

from .settings import DATABASES  # noqa
from .settings import *  # noqa

# pylint: skip-file

del DATABASES

TOX_ENV_NAME = os.environ["TOX_ENV_NAME"]
py_version, dj_version, pg_version = TOX_ENV_NAME.split("-")

if pg_version == "postgres12":
    from .settings_postgres12 import *  # noqa
elif pg_version == "postgres13":
    from .settings_postgres13 import *  # noqa
elif pg_version == "postgres14":
    from .settings_postgres14 import *  # noqa
elif pg_version == "postgres15":
    from .settings_postgres15 import *  # noqa
else:
    raise NotImplementedError(f"Invalid pg_version={pg_version} - TOX_ENV_NAME={TOX_ENV_NAME}")



from uuid import uuid4

import pytest
from app.models import Customer
from django.db import IntegrityError
from django.db import connection
from django.db import transaction

from django_partitioned_audit.audit_table.advanced_audit_table_manager import AdvancedAuditTableManager
from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager
from django_partitioned_audit.audit_table.simple_audit_table_manager import SimpleAuditTableManager
from django_partitioned_audit.models import AdvancedAuditEntry
from django_partitioned_audit.models import SimpleAuditEntry
from django_partitioned_audit.partitions.partition_manager_time_range import PartitionManager
from django_partitioned_audit.partitions.time_range_partitioning import DailyTimeRangeGenerator

# pylint: skip-file


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize(
    "audit_table_manager_class,model_class",
    [[SimpleAuditTableManager, SimpleAuditEntry], [AdvancedAuditTableManager, AdvancedAuditEntry]],
)
def test_audit_table_manager(audit_table_manager_class, model_class, settings_audit_table_manager, remove_auditing):
    # TODO: we're testing too much here, need to refactor and split this in many tests
    settings_audit_table_manager(audit_table_manager_class)
    audit_manager = AuditTableManager.get_implementation()
    assert isinstance(audit_manager, audit_table_manager_class)

    customer_count = 0

    # >>> let's create an instance without auditing, it should work
    assert Customer.objects.all().count() == customer_count
    Customer.objects.create(name=str(uuid4()))  # make sure model without auditing works
    customer_count += 1
    assert Customer.objects.all().count() == customer_count

    # >>> create partitioned table
    pm = PartitionManager(audit_manager=audit_manager, time_range_generator=DailyTimeRangeGenerator())
    assert audit_manager.partitioned_table_name not in pm.get_tables()
    with connection.cursor() as cursor:
        audit_manager.create_partitioned_table(cursor)
    assert audit_manager.partitioned_table_name in pm.get_tables()

    # >>> let's create an instance without auditing, it should work
    Customer.objects.create(name=str(uuid4()))  # make sure model without auditing works
    customer_count += 1
    assert Customer.objects.all().count() == customer_count

    # >>> enable auditing
    with connection.cursor() as cursor:
        audit_manager.enable_audit_for_table(cursor, Customer._meta.db_table)

    # >>> partitioned table exists, but no partitions... insert should fail
    with transaction.atomic():
        with pytest.raises(IntegrityError, match=r"no partition of relation .* found for row"):
            Customer.objects.create(name=str(uuid4()))

    # >>> let's create the partition
    pi = pm.generate_partition_info_for_current_period()
    sql, params = audit_manager.generate_create_partition(pi.partition, pi.partitioned_table, pi.from_date, pi.to_date)
    with connection.cursor() as cursor:
        cursor.execute(sql, params)

    # >>> let's create a model & test audit entries
    assert model_class.objects.all().count() == 0
    customer = Customer.objects.create(name=str(uuid4()))
    customer_count += 1
    assert Customer.objects.all().count() == customer_count
    assert model_class.objects.all().count() == 1
    customer.name = str(uuid4())
    customer.save()
    assert model_class.objects.all().count() == 2



from uuid import uuid4

import pytest
from app import models as models_first_app
from django.db import IntegrityError
from django.db import ProgrammingError
from django.db import transaction
from second_app import models as models_second_app

from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager

# pylint: skip-file


@pytest.mark.django_db(transaction=True)
def test_without_auditing(truncate_partitions, remove_auditing):
    ModelClass = AuditTableManager.get_implementation().model  # noqa

    customer_first_app = models_first_app.Customer.objects.create(pk=4321, name=str(uuid4()))
    customer_second_app = models_second_app.Customer.objects.create(pk=4321, name=str(uuid4()))

    with pytest.raises(IntegrityError):
        # inserting again with same PK should fail
        models_first_app.Customer.objects.create(pk=4321, name=str(uuid4()))

    with pytest.raises(IntegrityError):
        # inserting again with same PK should fail
        models_second_app.Customer.objects.create(pk=4321, name=str(uuid4()))

    assert models_first_app.Customer.objects.filter(name=customer_first_app.name).get().pk == 4321
    assert models_second_app.Customer.objects.filter(name=customer_second_app.name).get().pk == 4321

    with transaction.atomic():
        with pytest.raises(ProgrammingError, match=r"relation .* does not exist"):
            ModelClass.objects.filter_model(models_first_app.Customer).count()


@pytest.mark.django_db(transaction=True)
def test_audit_two_app_with_same_model_names(partition_created, post_test_disable_trigger):
    ModelClass = AuditTableManager.get_implementation().model  # noqa

    assert ModelClass.objects.filter_model(models_first_app.Customer).count() == 0
    assert ModelClass.objects.filter_model(models_second_app.Customer).count() == 0

    customer_first_app = models_first_app.Customer.objects.create(pk=4321, name=str(uuid4()))
    assert ModelClass.objects.filter_model(models_first_app.Customer).count() == 1
    assert ModelClass.objects.filter_model(models_second_app.Customer).count() == 0

    customer_second_app = models_second_app.Customer.objects.create(pk=4321, name=str(uuid4()))
    assert ModelClass.objects.filter_model(models_first_app.Customer).count() == 1
    assert ModelClass.objects.filter_model(models_second_app.Customer).count() == 1

    audit = ModelClass.objects.filter_model(models_first_app.Customer).get()
    assert audit.get_row_data()["name"] == customer_first_app.name

    audit = ModelClass.objects.filter_model(models_second_app.Customer).get()
    assert audit.get_row_data()["name"] == customer_second_app.name



import time

from app.models import Customer
from app.models import Invoice
from app.models import Product
from django.http import HttpRequest
from django.urls import reverse_lazy
from django.views.generic import CreateView
from django.views.generic import DeleteView
from django.views.generic import UpdateView


class CustomerCreate(CreateView):
    success_url = reverse_lazy("customer/create")
    model = Customer
    fields = "__all__"


class CustomerUpdate(UpdateView):
    success_url = reverse_lazy("customer/create")
    model = Customer
    fields = "__all__"

    def post(self, request: HttpRequest, *args, **kwargs):
        pre_wait_ms = int(request.GET.get("pre-wait-ms", "0"))
        post_wait_ms = int(request.GET.get("post-wait-ms", "0"))
        if pre_wait_ms > 0:
            time.sleep(pre_wait_ms / 1000.0)
        response = super().post(request, *args, **kwargs)
        if post_wait_ms > 0:
            time.sleep(post_wait_ms / 1000.0)
        return response


class CustomerDelete(DeleteView):
    success_url = reverse_lazy("customer/create")
    model = Customer
    fields = "__all__"


class InvoiceCreate(CreateView):
    success_url = reverse_lazy("invoice/create")
    model = Invoice
    fields = "__all__"


class ProductCreate(CreateView):
    success_url = reverse_lazy("product/create")
    model = Product
    fields = "__all__"



import dataclasses
import datetime
import re
from typing import List
from typing import Optional
from typing import Tuple

PARTITION_PARSER = re.compile(
    r"""
    ^
    (?P<partitioned_table>.+)          # table name prefix
    _                         # separator
    (?P<from_year>\d{4})      # from year
    (?P<from_month>\d{2})     # from month
    (?P<from_day>\d{2})       # from day
    _                         # separator
    (?P<to_year>\d{4})        # to year
    (?P<to_month>\d{2})       # to month
    (?P<to_day>\d{2})         # to day
    $
""",
    re.VERBOSE,
)


@dataclasses.dataclass(frozen=True)
class PartitionInfo:
    partition: str  # was: `table_name`
    from_date: datetime.date
    to_date: datetime.date
    partitioned_table: str

    @classmethod
    def create(cls, partitioned_table: str, from_date: datetime.date, to_date: datetime.date) -> "PartitionInfo":
        partition = f"{partitioned_table}_{from_date.strftime('%Y%m%d')}_{to_date.strftime('%Y%m%d')}"
        return PartitionInfo(
            partition=partition, from_date=from_date, to_date=to_date, partitioned_table=partitioned_table
        )

    @classmethod
    def parse(cls, table_name: str) -> Optional["PartitionInfo"]:
        match = PARTITION_PARSER.fullmatch(table_name)
        if not match:
            return None

        from_date = datetime.date(
            year=int(match.group("from_year")), month=int(match.group("from_month")), day=int(match.group("from_day"))
        )
        to_date = datetime.date(
            year=int(match.group("to_year")), month=int(match.group("to_month")), day=int(match.group("to_day"))
        )
        if from_date >= to_date:
            raise ValueError(f"from_date {from_date} must be before to_date {to_date}")
        return PartitionInfo(
            partition=table_name,
            partitioned_table=match.group("partitioned_table"),
            from_date=from_date,
            to_date=to_date,
        )

    def generate_create_partition(self) -> Tuple[str, List[object]]:
        from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager

        return AuditTableManager.get_implementation().generate_create_partition(
            self.partition, self.partitioned_table, self.from_date, self.to_date
        )


@dataclasses.dataclass
class TimeRange:
    from_date: datetime.date
    to_date: datetime.date

    def __post_init__(self):
        assert self.from_date < self.to_date



from typing import List

from django.apps import AppConfig


def get_enabled_models(app_config: AppConfig) -> List[str]:
    try:
        trigger_audit_models = app_config.trigger_audit_models
    except AttributeError:
        return []
    return trigger_audit_models


def get_disabled_models(app_config: AppConfig) -> List[str]:
    try:
        trigger_audit_models = app_config.audit_disabled
    except AttributeError:
        return []
    return trigger_audit_models



from app import views
from django.contrib import admin
from django.urls import path

urlpatterns = [
    path("admin/", admin.site.urls),
    path("customer/create/", views.CustomerCreate.as_view(), name="customer/create"),
    path("customer/<int:pk>/update/", views.CustomerUpdate.as_view(), name="customer/update"),
    path("customer/<int:pk>/delete/", views.CustomerDelete.as_view(), name="customer/delete"),
    path("invoice/create/", views.InvoiceCreate.as_view(), name="invoice/create"),
    path("product/create/", views.ProductCreate.as_view(), name="product/create"),
]



from unittest.mock import Mock
from unittest.mock import patch
from uuid import uuid4

import pytest

from django_partitioned_audit.partitions.partition_manager_time_range import PartitionManager
from django_partitioned_audit.partitions.time_range_partitioning import MonthlyTimeRangeGenerator as MTRG
from tests.dummies import DummyAuditTableManager


@pytest.mark.django_db(transaction=True)
def test_short_name_works(partitioned_table: str):
    pm = PartitionManager(time_range_generator=MTRG())
    assert partitioned_table in pm.get_tables()


@pytest.mark.django_db(transaction=True)
def test_long_name_is_truncated(very_long_named_partitioned_table: str):
    with patch.object(PartitionManager, "_validate_partitioned_table", new_callable=Mock) as mock:
        mock.return_value = None
        pm = PartitionManager(
            audit_manager=DummyAuditTableManager(very_long_named_partitioned_table), time_range_generator=MTRG()
        )
        assert very_long_named_partitioned_table not in pm.get_tables()
        assert very_long_named_partitioned_table[0:63] in pm.get_tables()


@pytest.mark.django_db(transaction=True)
def test_partition_manager_fails_with_long_names():
    # "abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcdefghi_abc"
    # "abcdefghi_abcdefghi_abcdefghi_abcdefghi_abcde_19990101_19990101"
    valid_partitioned_table = f"abcdefghi_{uuid4().hex[0:9]}_{uuid4().hex[0:9]}_{uuid4().hex[0:9]}_abcde"
    invalid_partitioned_table = f"abcdefghi_{uuid4().hex[0:9]}_{uuid4().hex[0:9]}_{uuid4().hex[0:9]}_abcdef"
    partition_suffix = "_19990101_19990201"

    assert len(valid_partitioned_table) == 45
    assert len(invalid_partitioned_table) == 46
    assert len(partition_suffix) == 18

    assert PartitionManager(audit_manager=DummyAuditTableManager(valid_partitioned_table), time_range_generator=MTRG())
    with pytest.raises(ValueError):
        assert PartitionManager(
            audit_manager=DummyAuditTableManager(invalid_partitioned_table), time_range_generator=MTRG()
        )



import os
import pathlib
import re
from datetime import date
from typing import List
from typing import Tuple
from typing import Type

from django.db import ProgrammingError
from django.db import transaction
from django.db.models import Model

from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager
from django_partitioned_audit.models import AdvancedAuditEntry


class AdvancedAuditTableManager(AuditTableManager):
    """
    Audit table implementation based on:
    - https://wiki.postgresql.org/wiki/Audit_trigger_91plus
    - https://github.com/2ndQuadrant/audit-trigger
    """

    @property
    def partitioned_table_name(self) -> str:
        return "logged_actions"

    @property
    def model(self) -> Type[Model]:
        return AdvancedAuditEntry

    def generate_create_partition(
        self, partition_name: str, partitioned_table: str, from_date: date, to_date: date
    ) -> Tuple[str, List[object]]:
        # TODO: find a way to escape tables names, so we don't have to use that regex
        sql_statement = f"""
        CREATE TABLE {partition_name}
            PARTITION OF {partitioned_table}
            FOR VALUES FROM (%s) TO (%s);
        """
        params = [from_date, to_date]
        return sql_statement, params

    def create_partitioned_table(self, cursor):
        sql = pathlib.Path(__file__).parent.resolve() / "advanced_audit_table_manager.sql"
        assert sql.exists()
        try:
            with transaction.atomic():
                cursor.execute(f"SELECT * FROM {self.partitioned_table_name} limit 1")
            audit_table_exists = True
        except ProgrammingError as err:
            audit_table_exists = None
            if err.args[0]:
                if isinstance(err.args[0], str):
                    message = err.args[0].splitlines()[0]
                    if re.search(rf"relation.*{self.partitioned_table_name}.*does not exist", message):
                        audit_table_exists = False
            if audit_table_exists is None:
                raise

        if not audit_table_exists:
            cursor.execute(sql.read_text())
            cursor.execute(
                """
            CREATE OR REPLACE VIEW logged_actions_view AS (
                SELECT
                    logged_actions.*,
                    hstore_to_jsonb(logged_actions.row_data) as "row_data_json",
                    hstore_to_jsonb(logged_actions.changed_fields) as "changed_fields_json",
                    ROW_NUMBER() OVER () as "synthetic_id"
                FROM
                    logged_actions
            );
            """
            )
            # TODO: remove `synthetic_id` (only reason to have it is to make the view read-only)

    def enable_audit_for_table(self, cursor, table_name: str):
        sql = "SELECT audit_table(%s);"
        cursor.execute(sql, [table_name])

    def disable_audit_for_table(self, cursor, table_name: str):
        cursor.execute(
            "DROP TRIGGER IF EXISTS audit_trigger_row ON {table_name};".format(
                table_name=table_name,  # FIXME: properly escape this to guarantee avoid SQL injection
            )
        )
        cursor.execute(
            "DROP TRIGGER IF EXISTS audit_trigger_stm ON {table_name};".format(
                table_name=table_name,  # FIXME: properly escape this to guarantee avoid SQL injection
            )
        )

    def drop_all(self, cursor):
        assert "PYTEST_CURRENT_TEST" in os.environ
        for table_name, trigger_name in self.get_triggers(cursor):
            if trigger_name in ["audit_trigger_row", "audit_trigger_stm"]:
                cursor.execute(f"DROP TRIGGER IF EXISTS {trigger_name} ON {table_name}")

        drop_partitions = """
        DO
        $src$
        DECLARE
            item VARCHAR;
        BEGIN
            FOR item IN
                SELECT table_name FROM information_schema.tables WHERE table_name LIKE %s
            LOOP
                PERFORM 'DROP TABLE IF EXISTS ' || item;
            END LOOP;
        END;
        $src$
        """
        params = [f"{self.partitioned_table_name}_????????_????????"]
        cursor.execute(drop_partitions, params)

        cursor.execute(f"DROP VIEW IF EXISTS {self.partitioned_table_name}_view")
        cursor.execute(f"DROP TABLE IF EXISTS {self.partitioned_table_name}")
        return



import os
from datetime import date
from io import StringIO
from typing import List
from unittest.mock import PropertyMock
from unittest.mock import patch

import pytest
from django.core import management

from django_partitioned_audit.partitions.partition_manager_time_range import PartitionManager

# pylint: skip-file


def print_unindented(text: str):
    print("\n".join([line.strip() for line in text.splitlines()]))


def run_command(*command_args):
    with patch.object(PartitionManager, "today", new_callable=PropertyMock) as mock_today:
        mock_today.return_value = date(2022, 2, 22)
        out = StringIO()
        management.call_command(
            "manage_partition_tables",
            *command_args,
            verbosity=0,
            stdout=out,
        )
        lines: List[str] = out.getvalue().splitlines()
        print("")
        print(f"    $ manage.py manage_partition_tables {' '.join(command_args)}")
        for line in lines:
            print(f"    {line}")
        print("")


@pytest.mark.skipif(os.environ.get("TEST_README") != "1", reason="TEST_README")
@pytest.mark.django_db(transaction=True)
def test_simple_create(truncate_partitions):
    run_command("create", "--extra-days=60")


@pytest.mark.skipif(os.environ.get("TEST_README") != "1", reason="TEST_README")
@pytest.mark.django_db(transaction=True)
def test_sample_workflow(truncate_partitions):
    print_unindented(
        """
    Sample usage
    ++++++++++++

    If you want to have enough partition to handle next 90 days (around 3 months), you can use `--extra-days=90`.
    Because it's the first time we run the command, no partition exists, and the plan will report that all
    partitions need to be created::
    """
    )
    run_command("simulate", "--extra-days=90")
    print_unindented(
        """
    We can also see the plan if no extra days are requested (this way, we'll only create partitions for
    the current month::
    """
    )
    run_command("simulate", "--extra-days=0")
    print_unindented(
        """
    Now let's create the partitions::
    """
    )
    run_command("create", "--extra-days=0")
    print_unindented(
        """
    If we run the command and we pass `--extra-days=90`, the partition for the current month already exists, and
    only partitions for next months (to cover 90 days) will be created::
    """
    )
    run_command("create", "--extra-days=90")
    print_unindented(
        """
    We can use `list` to list existing partitions::
    """
    )
    run_command("list")


@pytest.mark.skipif(os.environ.get("TEST_README") != "1", reason="TEST_README")
@pytest.mark.django_db(transaction=True)
def test_sample_workflow_weekly(truncate_partitions):
    print_unindented(
        """
    Partition per week
    ++++++++++++++++++

    We can use one partition per week::
    """
    )
    run_command("create", "--extra-days=30", "--time-range-generator=WeeklyTimeRangeGenerator")


@pytest.mark.skipif(os.environ.get("TEST_README") != "1", reason="TEST_README")
@pytest.mark.django_db(transaction=True)
def test_sample_workflow_daily(truncate_partitions):
    print_unindented(
        """
    Partition per day
    +++++++++++++++++

    We can use one partition per day::
    """
    )
    run_command("create", "--extra-days=10", "--time-range-generator=DailyTimeRangeGenerator")



import re
from datetime import date
from io import StringIO
from typing import List
from unittest.mock import Mock
from unittest.mock import PropertyMock
from unittest.mock import patch

import pytest
from django.core import management

from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager
from django_partitioned_audit.partitions.partition_manager_time_range import PartitionManager

# pylint: skip-file


@pytest.mark.django_db(transaction=True)
def test_flush():
    management.call_command("flush", verbosity=0, interactive=False)


@pytest.mark.django_db(transaction=True)
def test_migrate():
    management.call_command("migrate", verbosity=0, interactive=False)


@pytest.mark.django_db(transaction=True)
def test_manage_partition_tables_create_monthly(truncate_partitions, audit_table_manager: AuditTableManager):
    partitioned_table = audit_table_manager.partitioned_table_name
    with patch.object(PartitionManager, "get_existing_partitions", new_callable=Mock) as mock_partitions:
        with patch.object(PartitionManager, "today", new_callable=PropertyMock) as mock_today:
            mock_partitions.return_value = []
            mock_today.return_value = date(1877, 3, 22)
            out = StringIO()
            management.call_command("manage_partition_tables", "create", verbosity=0, stdout=out)
            lines: List[str] = out.getvalue().splitlines()
            assert lines

            lines_tables = [_ for _ in lines if re.search(rf"{partitioned_table}.*to create", _)]
            assert len(lines_tables) == 2
            assert any([re.search(rf"{partitioned_table}_18770301_18770401.*to create", _) for _ in lines_tables])
            assert any([re.search(rf"{partitioned_table}_18770401_18770501.*to create", _) for _ in lines_tables])


@pytest.mark.django_db(transaction=True)
def test_manage_partition_tables_create_weekly(truncate_partitions, audit_table_manager: AuditTableManager):
    partitioned_table = audit_table_manager.partitioned_table_name

    with patch.object(PartitionManager, "today", new_callable=PropertyMock) as mock_today:
        mock_today.return_value = date(1877, 3, 22)

        out = StringIO()
        management.call_command(
            "manage_partition_tables",
            "create",
            "--time-range-generator=WeeklyTimeRangeGenerator",
            verbosity=0,
            stdout=out,
        )
        lines: List[str] = out.getvalue().splitlines()
        assert lines

        lines_tables = [_ for _ in lines if re.search(rf"{partitioned_table}.*to create", _)]
        assert len(lines_tables) == 5
        assert any([re.search(rf"{partitioned_table}_18770322_18770329.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770329_18770405.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770405_18770412.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770412_18770419.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770419_18770426.*to create", _) for _ in lines_tables])


@pytest.mark.django_db(transaction=True)
def test_manage_partition_tables_create_daily(truncate_partitions, audit_table_manager: AuditTableManager):
    partitioned_table = audit_table_manager.partitioned_table_name

    with patch.object(PartitionManager, "today", new_callable=PropertyMock) as mock_today:
        mock_today.return_value = date(1877, 3, 22)

        out = StringIO()
        management.call_command(
            "manage_partition_tables",
            "create",
            "--time-range-generator=DailyTimeRangeGenerator",
            "--extra-days=6",
            verbosity=0,
            stdout=out,
        )
        lines: List[str] = out.getvalue().splitlines()
        assert lines

        lines_tables = [_ for _ in lines if re.search(rf"{partitioned_table}.*to create", _)]
        assert len(lines_tables) == 7
        assert any([re.search(rf"{partitioned_table}_18770322_18770323.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770323_18770324.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770324_18770325.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770325_18770326.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770326_18770327.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770327_18770328.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770328_18770329.*to create", _) for _ in lines_tables])


@pytest.mark.django_db(transaction=True)
def test_manage_partition_tables_create_twice(truncate_partitions, audit_table_manager: AuditTableManager):
    partitioned_table = audit_table_manager.partitioned_table_name

    with patch.object(PartitionManager, "today", new_callable=PropertyMock) as mock_today:
        mock_today.return_value = date(1877, 3, 22)

        out = StringIO()
        management.call_command("manage_partition_tables", "create", verbosity=0, stdout=out)
        lines: List[str] = out.getvalue().splitlines()
        assert lines

        lines_tables = [_ for _ in lines if re.search(rf"{partitioned_table}.*to create", _)]
        assert len(lines_tables) == 2
        assert any([re.search(rf"{partitioned_table}_18770301_18770401.*to create", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770401_18770501.*to create", _) for _ in lines_tables])

        out = StringIO()
        management.call_command("manage_partition_tables", "create", verbosity=0, stdout=out)
        lines: List[str] = out.getvalue().splitlines()
        assert lines

        lines_tables = [_ for _ in lines if re.search(rf"{partitioned_table}.*to create", _)]
        assert len(lines_tables) == 0

        lines_tables = [_ for _ in lines if re.search(rf"{partitioned_table}.*exists", _)]
        assert len(lines_tables) == 2
        assert any([re.search(rf"{partitioned_table}_18770301_18770401.*exists", _) for _ in lines_tables])
        assert any([re.search(rf"{partitioned_table}_18770401_18770501.*exists", _) for _ in lines_tables])



from datetime import date
from typing import Type
from unittest.mock import Mock
from unittest.mock import PropertyMock
from unittest.mock import patch

import pytest

from django_partitioned_audit.partitions.partition_info import PartitionInfo
from django_partitioned_audit.partitions.partition_manager_plan import Plan
from django_partitioned_audit.partitions.partition_manager_plan import Status
from django_partitioned_audit.partitions.partition_manager_time_range import PartitionManager
from django_partitioned_audit.partitions.time_range_partitioning import DailyTimeRangeGenerator
from django_partitioned_audit.partitions.time_range_partitioning import MonthlyTimeRangeGenerator
from django_partitioned_audit.partitions.time_range_partitioning import TimeRangeGenerator
from django_partitioned_audit.partitions.time_range_partitioning import WeeklyTimeRangeGenerator
from tests.conftest import skipifadvanced


class MonthlyScenario:
    prev2: PartitionInfo = PartitionInfo.parse("audit_simple_18770101_18770201")
    prev1: PartitionInfo = PartitionInfo.parse("audit_simple_18770201_18770301")
    cur: PartitionInfo = PartitionInfo.parse("audit_simple_18770301_18770401")
    next1: PartitionInfo = PartitionInfo.parse("audit_simple_18770401_18770501")
    next2: PartitionInfo = PartitionInfo.parse("audit_simple_18770501_18770601")
    extra_days: int = 45
    extra_days_current_time_range: int = 9  # `today=1877-03-22`, 9 days left in this month
    tmg_impl: Type[TimeRangeGenerator] = MonthlyTimeRangeGenerator


class WeeklyScenario:
    prev2: PartitionInfo = PartitionInfo.parse("audit_simple_18770308_18770315")
    prev1: PartitionInfo = PartitionInfo.parse("audit_simple_18770315_18770322")
    cur: PartitionInfo = PartitionInfo.parse("audit_simple_18770322_18770329")
    next1: PartitionInfo = PartitionInfo.parse("audit_simple_18770329_18770405")
    next2: PartitionInfo = PartitionInfo.parse("audit_simple_18770405_18770412")
    extra_days: int = 18
    extra_days_current_time_range: int = 6
    tmg_impl: Type[TimeRangeGenerator] = WeeklyTimeRangeGenerator


class DailyScenario:
    prev2: PartitionInfo = PartitionInfo.parse("audit_simple_18770320_18770321")
    prev1: PartitionInfo = PartitionInfo.parse("audit_simple_18770321_18770322")
    cur: PartitionInfo = PartitionInfo.parse("audit_simple_18770322_18770323")
    next1: PartitionInfo = PartitionInfo.parse("audit_simple_18770323_18770324")
    next2: PartitionInfo = PartitionInfo.parse("audit_simple_18770324_18770325")
    extra_days: int = 2
    extra_days_current_time_range: int = 0
    tmg_impl: Type[TimeRangeGenerator] = DailyTimeRangeGenerator


def generate_scenarios():
    for scenario in [MonthlyScenario, WeeklyScenario, DailyScenario]:
        # partitions for current period is missing
        yield (
            "part-cur-time-range-missing-1",
            [],
            0,
            Status.MISSING_CURRENT,
            [scenario.cur],
            scenario.tmg_impl,
            # let's force expansion
        )
        yield (
            "part-cur-time-range-missing-2",
            [scenario.prev2, scenario.prev1],
            0,
            Status.MISSING_CURRENT,
            [scenario.cur],
            scenario.tmg_impl,
        )
        yield (
            "part-cur-time-range-missing-3",
            [],
            scenario.extra_days,
            Status.MISSING_CURRENT,
            [scenario.cur, scenario.next1, scenario.next2],
            scenario.tmg_impl,
        )
        yield (
            "part-cur-time-range-missing-4",
            [scenario.prev2, scenario.prev1],
            scenario.extra_days,
            Status.MISSING_CURRENT,
            [scenario.cur, scenario.next1, scenario.next2],
            scenario.tmg_impl,
        )
        # partitions for next periods is missing
        yield (
            "cur-ok-but-extra-days-missing-1",
            [scenario.cur],
            scenario.extra_days,
            Status.MISSING_NEXT,
            [scenario.next1, scenario.next2],
            scenario.tmg_impl,
        )
        yield (
            "cur-ok-but-extra-days-missing-2",
            [scenario.prev2, scenario.prev1, scenario.cur],
            scenario.extra_days,
            Status.MISSING_NEXT,
            [scenario.next1, scenario.next2],
            scenario.tmg_impl,
        )
        # all good
        yield (
            "all-good-1",
            [scenario.cur],
            0,
            Status.OK,
            [],
            scenario.tmg_impl,
        )
        yield (
            "all-good-2",
            [scenario.cur],
            scenario.extra_days_current_time_range,
            Status.OK,
            [],
            scenario.tmg_impl,
        )
        yield (
            "all-good-3",
            [scenario.cur, scenario.next1, scenario.next2],
            scenario.extra_days,
            Status.OK,
            [],
            scenario.tmg_impl,
        )
        yield (
            "all-good-4",
            [scenario.prev2, scenario.prev1, scenario.cur, scenario.next1, scenario.next2],
            scenario.extra_days,
            Status.OK,
            [],
            scenario.tmg_impl,
        )


scenarios = list(generate_scenarios())


@skipifadvanced
@pytest.mark.parametrize("desc,existing_partitions,extra_days,exp_status,exp_new,trg_class", scenarios)
def test_simulate_scenarios(desc, existing_partitions, extra_days, exp_status, exp_new, trg_class):
    assert all([isinstance(_, PartitionInfo) for _ in existing_partitions])
    assert isinstance(extra_days, int)
    assert 0 <= extra_days <= 365
    assert exp_status in {Status.OK, Status.MISSING_CURRENT, Status.MISSING_NEXT}
    assert all([isinstance(_, PartitionInfo) for _ in exp_new])

    pm = PartitionManager(time_range_generator=trg_class(), extra_days=extra_days)
    with patch.object(PartitionManager, "get_existing_partitions", new_callable=Mock) as mock_partitions:
        with patch.object(PartitionManager, "today", new_callable=PropertyMock) as mock_today:
            mock_partitions.return_value = existing_partitions
            mock_today.return_value = date(1877, 3, 22)
            plan: Plan = pm.generate_plan()

            assert set(plan.existing_partitions) == set(existing_partitions)
            assert set(plan.partitions_to_create) == set(exp_new)
            assert plan.status is not None
            assert plan.status == exp_status



import abc
from datetime import date
from typing import List
from typing import Tuple
from typing import Type
from typing import TypeVar

from django.conf import settings
from django.db.models import Model
from django.utils.module_loading import import_string

SelfAuditTableManager = TypeVar("SelfAuditTableManager", bound="AuditTableManager")


DEFAULT_IMPL = "django_partitioned_audit.audit_table.simple_audit_table_manager.SimpleAuditTableManager"


class AuditTableManager(abc.ABC):
    @property
    @abc.abstractmethod
    def partitioned_table_name(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def model(self) -> Type[Model]:
        ...

    @abc.abstractmethod
    def create_partitioned_table(self, cursor):
        ...

    @abc.abstractmethod
    def enable_audit_for_table(self, cursor, table_name: str):
        ...

    @abc.abstractmethod
    def disable_audit_for_table(self, cursor, table_name: str):
        ...

    @abc.abstractmethod
    def generate_create_partition(
        self, partition_name: str, partitioned_table: str, from_date: date, to_date: date
    ) -> Tuple[str, List[object]]:
        ...

    @abc.abstractmethod
    def drop_all(self, cursor):
        ...

    @classmethod
    def get_implementation(cls) -> SelfAuditTableManager:
        try:
            impl_class_name = settings.DPA_AUDIT_TABLE_MANAGER
        except AttributeError:
            impl_class_name = DEFAULT_IMPL
        impl_class: Type[SelfAuditTableManager] = import_string(impl_class_name)
        return impl_class()

    def get_triggers(self, cursor) -> List[List[str]]:
        """Return the name of all triggers"""
        # https://www.postgresql.org/docs/current/infoschema-tables.html
        cursor.execute("SELECT event_object_table, trigger_name FROM information_schema.triggers")
        table_trigger_list = [[row[0], row[1]] for row in cursor.fetchall()]
        return table_trigger_list



from datetime import date
from typing import List
from typing import Union
from unittest.mock import Mock
from unittest.mock import PropertyMock
from unittest.mock import patch

import pytest

from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager
from django_partitioned_audit.partitions.partition_info import PartitionInfo
from django_partitioned_audit.partitions.partition_manager_time_range import PartitionManager
from django_partitioned_audit.partitions.time_range_partitioning import MonthlyTimeRangeGenerator


def test_generate_partition_info_for_current_period(audit_table_manager: AuditTableManager):
    pm = PartitionManager(time_range_generator=MonthlyTimeRangeGenerator())
    with patch.object(PartitionManager, "today", new_callable=PropertyMock) as mock:
        mock.return_value = date(1988, 3, 22)
        partition_info = pm.generate_partition_info_for_current_period()
        assert partition_info.partition == f"{audit_table_manager.partitioned_table_name}_19880301_19880401"

        mock.return_value = date(1988, 12, 22)
        partition_info = pm.generate_partition_info_for_current_period()
        assert partition_info.partition == f"{audit_table_manager.partitioned_table_name}_19881201_19890101"


def test_new_partition():
    pm = PartitionManager(time_range_generator=MonthlyTimeRangeGenerator())
    previous = PartitionInfo.create("part", date(1991, 1, 1), date(1991, 2, 1))

    assert pm.new_partition(previous) == PartitionInfo.create("part", date(1991, 2, 1), date(1991, 3, 1))


def test_new_partition_with_previous_starting_off_period():
    pm = PartitionManager(time_range_generator=MonthlyTimeRangeGenerator())
    assert pm.new_partition(PartitionInfo.create("part", date(1991, 1, 22), date(1991, 2, 1))) == PartitionInfo.create(
        "part", date(1991, 2, 1), date(1991, 3, 1)
    )


def test_new_partition_with_previous_ending_off_period():
    pm = PartitionManager(time_range_generator=MonthlyTimeRangeGenerator())
    assert pm.new_partition(PartitionInfo.create("part", date(1991, 1, 1), date(1991, 1, 20))) == PartitionInfo.create(
        "part", date(1991, 1, 20), date(1991, 2, 1)
    )

    assert pm.new_partition(PartitionInfo.create("part", date(1991, 1, 1), date(1991, 1, 31))) == PartitionInfo.create(
        "part", date(1991, 1, 31), date(1991, 2, 1)
    )


@pytest.mark.parametrize("params", [0, 1, 2, 30, [31], [32], [45], [55]])
def test_generate_plan_when_no_partition_exists(params: Union[int, List[int]], audit_table_manager: AuditTableManager):
    if isinstance(params, int):
        extra_days = params
        expected_partitions_to_create = [
            PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 3, 1), date(1991, 4, 1)),
        ]
    else:
        extra_days = params[0]
        expected_partitions_to_create = [
            PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 3, 1), date(1991, 4, 1)),
            PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 4, 1), date(1991, 5, 1)),
        ]

    pm = PartitionManager(extra_days=extra_days, time_range_generator=MonthlyTimeRangeGenerator())
    with patch.object(PartitionManager, "today", new_callable=PropertyMock) as mock_today:
        mock_today.return_value = date(1991, 3, 1)
        with patch.object(PartitionManager, "get_existing_partitions", new_callable=Mock) as mock_part:
            mock_part.return_value = []

            plan = pm.generate_plan()
            assert plan.existing_partitions == []
            assert plan.partitions_to_create == expected_partitions_to_create


@pytest.mark.parametrize("params", [0, 1, 2, 30, [31], [32], [45], [55]])
def test_generate_plan_when_partition_exists(params: Union[int, List[int]], audit_table_manager: AuditTableManager):
    if isinstance(params, int):
        extra_days = params
        expected_created = []
    else:
        extra_days = params[0]
        expected_created = [
            PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 4, 1), date(1991, 5, 1)),
        ]

    pm = PartitionManager(extra_days=extra_days, time_range_generator=MonthlyTimeRangeGenerator())
    with patch.object(PartitionManager, "today", new_callable=PropertyMock) as mock_today:
        mock_today.return_value = date(1991, 3, 1)
        with patch.object(PartitionManager, "get_existing_partitions", new_callable=Mock) as mock_part:
            mock_part.return_value = [
                PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 3, 1), date(1991, 4, 1)),
            ]

            plan = pm.generate_plan()
            assert plan.existing_partitions == [
                PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 3, 1), date(1991, 4, 1)),
            ]
            assert plan.partitions_to_create == expected_created


@pytest.mark.parametrize("params", [0, 1, 16, [17], [30]])
def test_generate_plan_when_partition_exists_but_not_for_current(
    params: Union[int, List[int]], audit_table_manager: AuditTableManager
):
    if isinstance(params, int):
        extra_days = params
        expected_created = [
            PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 3, 1), date(1991, 4, 1)),
        ]
    else:
        extra_days = params[0]
        expected_created = [
            PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 3, 1), date(1991, 4, 1)),
            PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 4, 1), date(1991, 5, 1)),
        ]

    pm = PartitionManager(extra_days=extra_days, time_range_generator=MonthlyTimeRangeGenerator())
    with patch.object(PartitionManager, "today", new_callable=PropertyMock) as mock_today:
        mock_today.return_value = date(1991, 3, 15)
        with patch.object(PartitionManager, "get_existing_partitions", new_callable=Mock) as mock_part:
            mock_part.return_value = [
                PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 2, 1), date(1991, 3, 1)),
            ]

            plan = pm.generate_plan()
            assert plan.existing_partitions == [
                PartitionInfo.create(audit_table_manager.partitioned_table_name, date(1991, 2, 1), date(1991, 3, 1)),
            ]
            assert plan.partitions_to_create == expected_created



# Generated by Django 4.2.2 on 2023-06-25 09:50

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("app", "0002_invoice_product"),
    ]

    operations = [
        migrations.CreateModel(
            name="InvoiceItem",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=100)),
            ],
        ),
    ]



MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    # 'django.middleware.csrf.CsrfViewMiddleware',  # let's make testing easier
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

SECRET_KEY = "lr_ciiy@srbd$-js3jx3w7ps7(9ul+1*k_4)q2qcpd*(9j*!fn"

ALLOWED_HOSTS = []

ROOT_URLCONF = "config.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "config.wsgi.application"

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True

STATIC_URL = "/static/"



import enum
import logging
from typing import Type

from django.contrib.postgres.fields import HStoreField
from django.db import models
from django.db.models import Model
from django.db.models import Q

import django_partitioned_audit.signals  # noqa pylint: disable=unused-import

logger = logging.getLogger(__name__)


class Operation(enum.Enum):
    INSERT = 1
    UPDATE = 2
    DELETE = 3
    TRUNCATE = 4


# ----------------------------------------------------------------------------------------------------
# SimpleAuditTableManager
# ----------------------------------------------------------------------------------------------------


class SimpleAuditEntryQuerySet(models.QuerySet):
    def filter_model(self, model_class: Type[Model]):
        """Filter audit events by the model that generated the events"""
        return self.filter(object_table__iexact=model_class._meta.db_table)  # noqa

    def filter_payload(self, **kwargs):
        """
        Filter audit events by a value in the `payload` of the audit event, that means,
        the value of the column at the time the audit even was generated.
        """
        assert kwargs
        filter_kwargs = {}
        for key_name, key_value in kwargs.items():
            filter_kwargs[f"object_payload__{key_name}"] = key_value

        return self.filter(**filter_kwargs)

    def filter_operation(self, operation: Operation):
        if operation == Operation.INSERT:
            return self.filter(audit_operation="I")
        elif operation == Operation.UPDATE:
            return self.filter(audit_operation="U")
        elif operation == Operation.DELETE:
            return self.filter(audit_operation="D")
        elif operation == Operation.TRUNCATE:
            raise NotImplementedError("TRUNCATE is not supported by this implementation")
        else:
            raise ValueError(f"Invalid operation: {operation}")


class SimpleAuditEntry(Model):
    """
    This model exists to facilitate reading audit events.
    This is mapped to the view.
    No update or delete operations can be done.
    """

    object_table = models.CharField(max_length=63)
    object_payload = models.JSONField()

    audit_entry_created = models.DateTimeField()
    audit_entry_created_statement_timestamp = models.DateTimeField()
    audit_txid_current = models.BigIntegerField()
    audit_operation = models.CharField(max_length=32)

    objects = SimpleAuditEntryQuerySet.as_manager()

    class Meta:
        managed = False
        db_table = "audit_simple_view"

    def __str__(self):
        return f"{self.__class__.__name__} " f"{self.object_table} " f"{self.audit_operation}"

    def is_insert(self):
        return self.audit_operation == "I"

    def is_update(self):
        return self.audit_operation == "U"

    def is_delete(self):
        return self.audit_operation == "D"

    def get_row_data(self) -> dict:
        return self.object_payload


# ----------------------------------------------------------------------------------------------------
# AdvancedAuditTableManager
# ----------------------------------------------------------------------------------------------------


class AdvancedAuditEntryQuerySet(models.QuerySet):
    def filter_model(self, model_class: Type[Model]):
        """Filter audit events by the model that generated the events"""
        return self.filter(table_name__iexact=model_class._meta.db_table)  # noqa

    def filter_payload(self, **kwargs):
        assert kwargs
        qs = self
        for key_name, key_value in kwargs.items():
            filter_kwargs_1 = {
                f"changed_fields_json__{key_name}": key_value,
            }
            filter_kwargs_2 = {
                f"changed_fields_json__{key_name}__isnull": True,
                f"row_data_json__{key_name}": key_value,
            }
            qs = qs.filter(Q(**filter_kwargs_1) | Q(**filter_kwargs_2))

        return qs

    def filter_operation(self, operation: Operation):
        if operation == Operation.INSERT:
            return self.filter(action="I")
        elif operation == Operation.UPDATE:
            return self.filter(action="U")
        elif operation == Operation.DELETE:
            return self.filter(action="D")
        elif operation == Operation.TRUNCATE:
            return self.filter(action="T")
        else:
            raise ValueError(f"Invalid operation: {operation}")


class AdvancedAuditEntry(Model):
    event_id = models.BigAutoField(primary_key=True, help_text="Unique identifier for each auditable event")
    schema_name = models.TextField(help_text="Database schema audited table for this event is in")
    table_name = models.TextField(help_text="Non-schema-qualified table name of table event occured in")
    action_tstamp_tx = models.DateTimeField(
        help_text="Transaction start timestamp for tx in which audited event occurred"
    )
    action_tstamp_stm = models.DateTimeField(
        help_text="Statement start timestamp for tx in which audited event occurred"
    )
    action_tstamp_clk = models.DateTimeField(
        help_text="Wall clock time at which audited event''s trigger call occurred"
    )
    transaction_id = models.BigIntegerField(
        help_text="Identifier of transaction that made the change. May wrap, but unique paired with action_tstamp_tx."
    )
    action = models.CharField(max_length=1, help_text="Action type; I = insert, D = delete, U = update, T = truncate")
    row_data = HStoreField(
        help_text="Record value. Null for statement-level trigger. For INSERT this is the new tuple. "
        "For DELETE and UPDATE it is the old tuple."
    )
    changed_fields = HStoreField(
        help_text="New values of fields changed by UPDATE. Null except for row-level UPDATE events."
    )

    # relid IS 'Table OID. Changes with drop/create. Get with ''tablename''::regclass';
    # session_user_name IS 'Login / session user whose statement caused the audited event';
    # client_addr IS 'IP address of client that issued query. Null for unix domain socket.';
    # client_port IS 'Remote peer IP port address of client that issued query. Undefined for unix socket.';
    # client_query IS 'Top-level query that caused this auditable event. May be more than one statement.';
    # application_name IS 'Application name set when this audit event occurred. Can be changed in-session by client.';
    # statement_only IS '''t'' if audit event is from an FOR EACH STATEMENT trigger, ''f'' for FOR EACH ROW';

    row_data_json = models.JSONField(
        help_text="Record value. Null for statement-level trigger. For INSERT this is the new tuple. "
        "For DELETE and UPDATE it is the old tuple."
    )

    changed_fields_json = models.JSONField(
        help_text="New values of fields changed by UPDATE. Null except for row-level UPDATE events."
    )

    objects = AdvancedAuditEntryQuerySet.as_manager()

    class Meta:
        managed = False
        db_table = "logged_actions_view"

    def __str__(self):
        return f"{self.__class__.__name__} " f"{self.table_name} " f"{self.action}"

    def is_insert(self):
        return self.action == "I"

    def is_update(self):
        return self.action == "U"

    def is_delete(self):
        return self.action == "D"

    def is_truncate(self):
        return self.action == "T"

    def get_row_data(self) -> dict:
        row_data = {}
        row_data.update(self.row_data_json or {})
        row_data.update(self.changed_fields_json or {})
        return row_data



from datetime import timedelta

from django_partitioned_audit.partitions.partition_info import PartitionInfo
from django_partitioned_audit.partitions.partition_info import TimeRange
from django_partitioned_audit.partitions.partition_manager_base import BasePartitionManager
from django_partitioned_audit.partitions.partition_manager_plan import Plan


class PartitionManager(BasePartitionManager):
    def generate_partition_info_for_current_period(self) -> PartitionInfo:
        """When there are partitions, we need to start with the one for current period"""
        from_to = self._time_range_generator.generate_partition_info_for_current_period(self.today)
        return PartitionInfo.create(self._partitioned_table, from_to.from_date, from_to.to_date)

    def new_partition(self, previous: PartitionInfo) -> PartitionInfo:
        from_to = self._time_range_generator.new_partition(TimeRange(previous.from_date, previous.to_date))
        return PartitionInfo.create(previous.partitioned_table, from_to.from_date, from_to.to_date)

    def generate_plan(self) -> Plan:
        """
        Generate `PartitionInfo` required to have partitions for at least `extra_days` days.
        """
        last_day_to_cover = self.today + timedelta(days=self._extra_days)
        plan: Plan = Plan()

        plan.existing_partitions.extend(self.get_existing_partitions())
        if not plan.last_partition:
            # no partition exists, we need at least one for the algorithm to work
            plan.partitions_to_create.append(self.generate_partition_info_for_current_period())

        while not plan.covers_date(last_day_to_cover):
            plan.partitions_to_create.append(self.new_partition(previous=plan.last_partition))

        plan.calculate_status(today=self.today, extra_days=self._extra_days)

        return plan



from django.apps import AppConfig


class AppConfig(AppConfig):
    name = "app"
    trigger_audit_models = [
        "Customer",
        "Product",
        "Invoice",
    ]
    audit_disabled = [
        "InvoiceItem",
    ]



import itertools
import random
import uuid
from concurrent import futures
from concurrent.futures.thread import ThreadPoolExecutor

import requests
from app.models import Customer
from django.core.management.base import BaseCommand
from django.urls import reverse

from django_partitioned_audit.models import SimpleAuditEntry


class Command(BaseCommand):
    help = "Closes the specified poll for voting"

    def add_arguments(self, parser):
        parser.add_argument("--server-url", default="http://127.0.0.1:8011")
        parser.add_argument("--items", default=10000, type=int)
        parser.add_argument("--max-workers", default=20, type=int)

    def handle(self, *args, **options):
        live_server = options["server_url"]
        items = options["items"]
        max_workers = options["max_workers"]

        print(f"server_url={live_server}")
        print(f"items={items}")
        print(f"max_workers={max_workers}")

        original_name = str(uuid.uuid4())
        customer = Customer.objects.create(name=original_name)
        urls = [
            # Don't wait for most of the requests
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=0",
            # Wait in different ways
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=100&post-wait-ms=0",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=0&post-wait-ms=100",
            f"{live_server}{reverse('customer/update', args=[customer.pk])}?pre-wait-ms=100&post-wait-ms=100",
        ]
        random.shuffle(urls)

        def _update(url: str, new_name: str):
            response = requests.post(url, data=dict(name=new_name))
            response.raise_for_status()

        new_names = [str(uuid.uuid4()) for _ in range(items)]

        # We can use a with statement to ensure threads are cleaned up promptly
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            # Start the load operations and mark each future with its URL
            url_iterator = itertools.cycle(urls)
            future_to_url = {executor.submit(_update, next(url_iterator), new_name): new_name for new_name in new_names}
            count = 0
            for future in futures.as_completed(future_to_url):
                count += 1
                # new_name = future_to_url[future]
                try:
                    future.result()
                except Exception:
                    raise
                else:
                    print("+", end="")
                    if count % 100 == 0:
                        print("")

        print("Finished. Now checking...")

        # Assert model was created and audit exists
        assert SimpleAuditEntry.objects.all().count() == len(new_names)
        names_in_payload = [audit_entry.get_row_data()["name"] for audit_entry in SimpleAuditEntry.objects.all()]
        assert len(names_in_payload) == len(new_names)
        assert set(names_in_payload) == set(new_names)



# Generated by Django 4.2.2 on 2023-06-19 10:19

import uuid

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("app", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Invoice",
            fields=[
                ("id", models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="Product",
            fields=[
                ("product_id", models.BigAutoField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=100)),
            ],
        ),
    ]



import uuid

import pytest
import requests
from app.models import Invoice
from app.models import Product
from django.urls import reverse

from django_partitioned_audit.models import SimpleAuditEntry
from tests.conftest import skipifadvanced

# pylint: skip-file


@skipifadvanced
@pytest.mark.django_db(transaction=True)
def test_pk_named_product_id(live_server, partition_created):
    name = str(uuid.uuid4())
    response = requests.post(f"{live_server}{reverse('product/create')}", data=dict(name=name))
    response.raise_for_status()
    assert Product.objects.filter(name=name).exists()

    # Assert model was created and audit exists
    audit_entry: SimpleAuditEntry = SimpleAuditEntry.objects.all().get()
    assert audit_entry.is_insert()
    assert audit_entry.get_row_data()["product_id"]
    assert audit_entry.get_row_data()["name"] == name


@skipifadvanced
@pytest.mark.django_db(transaction=True)
def test_pk_is_uuid(live_server, partition_created):
    name = str(uuid.uuid4())
    response = requests.post(f"{live_server}{reverse('invoice/create')}", data=dict(name=name))
    response.raise_for_status()
    assert Invoice.objects.filter(name=name).exists()

    # Assert model was created and audit exists
    audit_entry: SimpleAuditEntry = SimpleAuditEntry.objects.all().get()
    assert audit_entry.is_insert()
    assert audit_entry.get_row_data()["id"]
    assert audit_entry.get_row_data()["name"] == name



import pytest
from django.db import connection

from django_partitioned_audit.audit_table.audit_table_manager import AuditTableManager
from django_partitioned_audit.partitions.partition_info import PartitionInfo
from django_partitioned_audit.partitions.partition_manager_time_range import BasePartitionManager
from django_partitioned_audit.partitions.partition_manager_time_range import PartitionManager
from django_partitioned_audit.partitions.time_range_partitioning import MonthlyTimeRangeGenerator as MTRG
from django_partitioned_audit.partitions.time_range_partitioning import NotImplementedTimeRangeGenerator
from tests.dummies import DummyAuditTableManager


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("tmg_class", [MTRG, NotImplementedTimeRangeGenerator])
def test_get_tables(tmg_class, audit_table_manager: AuditTableManager):
    pm = BasePartitionManager(time_range_generator=tmg_class())
    tables = pm.get_tables()
    assert set(tables) >= {
        "app_customer",
        "app_invoice",
        "app_invoiceitem",
        "app_product",
        "second_app_customer",
        "second_app_invoice",
        "second_app_product",
        "auth_group",
        "auth_group_permissions",
        "auth_permission",
        "auth_user",
        "auth_user_groups",
        "auth_user_user_permissions",
        "django_admin_log",
        "django_content_type",
        "django_migrations",
        "django_session",
        audit_table_manager.partitioned_table_name,
        f"{audit_table_manager.partitioned_table_name}_view",
    }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("tmg_class", [MTRG, NotImplementedTimeRangeGenerator])
def test_get_existing_partitions_when_partitions_exists(partitioned_table: str, tmg_class):
    with connection.cursor() as cursor:
        sql = f"""
        CREATE TABLE {partitioned_table}_19990101_19990201
            PARTITION OF {partitioned_table}
            FOR VALUES FROM ('1999-01-01') TO ('1999-02-01');
        """
        cursor.execute(sql)

    partition = f"{partitioned_table}_19990101_19990201"

    pm = PartitionManager(audit_manager=DummyAuditTableManager(partitioned_table), time_range_generator=tmg_class())
    assert partitioned_table in pm.get_tables()
    assert partition in pm.get_tables()

    assert set(pm.get_existing_partitions()) == {PartitionInfo.parse(partition)}


@pytest.mark.django_db
@pytest.mark.parametrize("tmg_class", [MTRG, NotImplementedTimeRangeGenerator])
def test_get_existing_partitions_when_no_partitions(partitioned_table: str, tmg_class):
    pm = PartitionManager(time_range_generator=tmg_class())
    assert partitioned_table in pm.get_tables()
    partitions = list(pm.get_existing_partitions())
    assert len(partitions) == 0


@pytest.mark.django_db
@pytest.mark.parametrize("tmg_class", [MTRG, NotImplementedTimeRangeGenerator])
def test_create_partitions(partitioned_table: str, tmg_class):
    pm = PartitionManager(audit_manager=DummyAuditTableManager(partitioned_table), time_range_generator=tmg_class())

    partition = f"{partitioned_table}_19990101_19990201"
    partition_info = PartitionInfo.parse(partition)

    assert partitioned_table in pm.get_tables()
    assert partition not in pm.get_tables()

    pm.create_partitions([partition_info])

    assert partition in pm.get_tables()
